import datetime as dt
import logging
import os
from textwrap import dedent
from time import sleep

import pytest
from freezegun import freeze_time

import khal.khalendar.exceptions
import khal.utils
from khal import icalendar as icalendar_helpers
from khal.controllers import human_formatter
from khal.khalendar import CalendarCollection
from khal.khalendar.backend import CouldNotCreateDbDir
from khal.khalendar.event import Event
from khal.khalendar.vdir import Item

from . import utils
from .utils import (
    BERLIN,
    LOCALE_BERLIN,
    LOCALE_SYDNEY,
    LONDON,
    SYDNEY,
    CollVdirType,
    DumbItem,
    _get_text,
    cal1,
    cal2,
    cal3,
    normalize_component,
)

today = dt.date.today()
yesterday = today - dt.timedelta(days=1)
tomorrow = today + dt.timedelta(days=1)

aday = dt.date(2014, 4, 9)
bday = dt.date(2014, 4, 10)

event_allday_template = """BEGIN:VEVENT
SEQUENCE:0
UID:uid3@host1.com
DTSTART;VALUE=DATE:{}
DTEND;VALUE=DATE:{}
SUMMARY:a meeting
DESCRIPTION:short description
LOCATION:LDB Lobby
END:VEVENT"""

event_today = event_allday_template.format(
    today.strftime('%Y%m%d'), tomorrow.strftime('%Y%m%d'))
item_today = Item(event_today)
SIMPLE_EVENT_UID = 'V042MJ8B3SJNFXQOJL6P53OFMHJE8Z3VZWOU'


class TestCalendar:

    def test_create(self, coll_vdirs):
        assert True

    def test_new_event(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        event = coll.create_event_from_ics(event_today, cal1)
        assert event.calendar == cal1
        coll.insert(event)
        events = list(coll.get_events_on(today))
        assert len(events) == 1
        assert events[0].color == 'dark blue'
        assert len(list(coll.get_events_on(tomorrow))) == 0
        assert len(list(coll.get_events_on(yesterday))) == 0
        assert len(list(vdirs[cal1].list())) == 1

    def test_sanity(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        mtimes = {}
        for _ in range(100):
            for cal in coll._calendars:
                mtime = coll._local_ctag(cal)
                if mtimes.get(cal):
                    assert mtimes[cal] == mtime
                else:
                    mtimes[cal] = mtime

    def test_db_needs_update(self, coll_vdirs, sleep_time):
        coll, vdirs = coll_vdirs

        print('init')
        for calendar in coll._calendars:
            print(f'{calendar}: saved ctag: {coll._local_ctag(calendar)}, '
                  f'vdir ctag: {coll._backend.get_ctag(calendar)}')
        assert len(list(vdirs[cal1].list())) == 0
        assert coll._needs_update(cal1) is False
        sleep(sleep_time)

        vdirs[cal1].upload(item_today)
        print('upload')
        for calendar in coll._calendars:
            print(f'{calendar}: saved ctag: {coll._local_ctag(calendar)}, '
                  f'vdir ctag: {coll._backend.get_ctag(calendar)}')
        assert len(list(vdirs[cal1].list())) == 1
        assert coll._needs_update(cal1) is True
        coll.update_db()
        print('updated')
        for calendar in coll._calendars:
            print(f'{calendar}: saved ctag: {coll._local_ctag(calendar)}, '
                  f'vdir ctag: {coll._backend.get_ctag(calendar)}')
        assert coll._needs_update(cal1) is False


class TestVdirsyncerCompat:
    def test_list(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        event = Event.fromString(_get_text('event_d'), calendar=cal1, locale=LOCALE_BERLIN)
        assert event.etag is None
        assert event.href is None
        coll.insert(event)
        assert event.etag is not None
        assert event.href == 'V042MJ8B3SJNFXQOJL6P53OFMHJE8Z3VZWOU.ics'
        event = Event.fromString(event_today, calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event)
        hrefs = sorted(href for href, etag in coll._backend.list(cal1))
        assert {str(coll.get_event(href, calendar=cal1).uid) for href in hrefs} == {
            'uid3@host1.com',
            'V042MJ8B3SJNFXQOJL6P53OFMHJE8Z3VZWOU',
        }


class TestCollection:

    astart = dt.datetime.combine(aday, dt.time.min)
    aend = dt.datetime.combine(aday, dt.time.max)
    bstart = dt.datetime.combine(bday, dt.time.min)
    bend = dt.datetime.combine(bday, dt.time.max)
    astart_berlin = utils.BERLIN.localize(astart)
    aend_berlin = utils.BERLIN.localize(aend)
    bstart_berlin = utils.BERLIN.localize(bstart)
    bend_berlin = utils.BERLIN.localize(bend)

    def test_default_calendar(self, tmpdir):
        calendars = {
            'foobar': {'name': 'foobar', 'path': str(tmpdir), 'readonly': True},
            'home': {'name': 'home', 'path': str(tmpdir)},
            "Dad's Calendar": {'name': "Dad's calendar", 'path': str(tmpdir), 'readonly': True},
        }
        coll = CalendarCollection(
            calendars=calendars, locale=LOCALE_BERLIN, dbpath=':memory:',
        )
        assert coll.default_calendar_name is None
        with pytest.raises(ValueError):
            coll.default_calendar_name = "Dad's calendar"
        assert coll.default_calendar_name is None
        with pytest.raises(ValueError):
            coll.default_calendar_name = 'unknownstuff'
        assert coll.default_calendar_name is None
        coll.default_calendar_name = 'home'
        assert coll.default_calendar_name == 'home'
        assert coll.writable_names == ['home']

    def test_empty(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        start = dt.datetime.combine(today, dt.time.min)
        end = dt.datetime.combine(today, dt.time.max)
        assert list(coll.get_floating(start, end)) == []
        assert list(coll.get_localized(utils.BERLIN.localize(start),
                                       utils.BERLIN.localize(end))) == []

    def test_insert(self, coll_vdirs):
        """insert a localized event"""
        coll, vdirs = coll_vdirs
        coll.insert(
            Event.fromString(_get_text('event_dt_simple'), calendar=cal1, locale=LOCALE_BERLIN),
            cal1)
        events = list(coll.get_localized(self.astart_berlin, self.aend_berlin))
        assert len(events) == 1
        assert events[0].color == 'dark blue'
        assert events[0].calendar == cal1

        events = list(coll.get_events_on(aday))
        assert len(events) == 1
        assert events[0].color == 'dark blue'
        assert events[0].calendar == cal1

        assert len(list(vdirs[cal1].list())) == 1
        assert len(list(vdirs[cal2].list())) == 0
        assert len(list(vdirs[cal3].list())) == 0
        assert list(coll.get_floating(self.astart, self.aend)) == []

    def test_insert_d(self, coll_vdirs):
        """insert a floating event"""
        coll, vdirs = coll_vdirs
        event = Event.fromString(_get_text('event_d'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        events = list(coll.get_events_on(aday))
        assert len(events) == 1
        assert events[0].calendar == cal1
        assert events[0].color == 'dark blue'
        assert len(list(vdirs[cal1].list())) == 1
        assert len(list(vdirs[cal2].list())) == 0
        assert len(list(vdirs[cal3].list())) == 0
        assert list(coll.get_localized(self.bstart_berlin, self.bend_berlin)) == []

    def test_insert_d_no_value(self, coll_vdirs):
        """insert a date event with no VALUE=DATE option"""
        coll, vdirs = coll_vdirs
        coll.insert(
            Event.fromString(
                _get_text('event_d_no_value'), calendar=cal1, locale=LOCALE_BERLIN),
            cal1)
        events = list(coll.get_events_on(aday))
        assert len(events) == 1
        assert events[0].calendar == cal1
        assert len(list(vdirs[cal1].list())) == 1
        assert len(list(vdirs[cal2].list())) == 0
        assert len(list(vdirs[cal3].list())) == 0
        assert list(coll.get_localized(self.bstart_berlin, self.bend_berlin)) == []

    def test_get(self, coll_vdirs):
        """test getting an event by its href"""
        coll, vdirs = coll_vdirs
        event = Event.fromString(
            _get_text('event_dt_simple'), href='xyz.ics', calendar=cal1, locale=LOCALE_BERLIN,
        )
        coll.insert(event, cal1)
        event_from_db = coll.get_event(SIMPLE_EVENT_UID + '.ics', cal1)
        with freeze_time('2016-1-1'):
            assert normalize_component(event_from_db.raw) == \
                normalize_component(_get_text('event_dt_simple_inkl_vtimezone'))
        assert event_from_db.etag

    def test_change(self, coll_vdirs):
        """moving an event from one calendar to another"""
        coll, vdirs = coll_vdirs
        event = Event.fromString(_get_text('event_dt_simple'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        event = list(coll.get_events_on(aday))[0]
        assert event.calendar == cal1

        coll.change_collection(event, cal2)
        events = list(coll.get_events_on(aday))
        assert len(events) == 1
        assert events[0].calendar == cal2

    def test_update_event(self, coll_vdirs):
        """updating one event"""
        coll, vdirs = coll_vdirs
        event = Event.fromString(
            _get_text('event_dt_simple'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        events = coll.get_events_on(aday)
        event = list(events)[0]
        event.update_summary('really simple event')
        event.update_start_end(bday, bday)
        coll.update(event)
        events = list(coll.get_localized(self.astart_berlin, self.aend_berlin))
        assert len(events) == 0
        events = list(coll.get_floating(self.bstart, self.bend))
        assert len(events) == 1
        assert events[0].summary == 'really simple event'

    def test_newevent(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        bday = dt.datetime.combine(aday, dt.time.min)
        anend = bday + dt.timedelta(hours=1)
        event = icalendar_helpers.new_vevent(
            dtstart=bday, dtend=anend, summary="hi", timezone=utils.BERLIN,
            locale=LOCALE_BERLIN,
        )
        event = coll.create_event_from_ics(event.to_ical(), coll.default_calendar_name)
        assert event.allday is False

    def test_modify_readonly_calendar(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        coll._calendars[cal1]['readonly'] = True
        coll._calendars[cal3]['readonly'] = True
        event = Event.fromString(_get_text('event_dt_simple'), calendar=cal1, locale=LOCALE_BERLIN)

        with pytest.raises(khal.khalendar.exceptions.ReadOnlyCalendarError):
            coll.insert(event, cal1)
        with pytest.raises(khal.khalendar.exceptions.ReadOnlyCalendarError):
            # params don't really matter here
            coll.delete('href', 'eteg', cal1)

    def test_search(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        assert len(list(coll.search('Event'))) == 0
        event = Event.fromString(
            _get_text('event_dt_simple'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        assert len(list(coll.search('Event'))) == 1
        event = Event.fromString(
            _get_text('event_dt_floating'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        assert len(list(coll.search('Search for me'))) == 1
        assert len(list(coll.search('Event'))) == 2

    def test_search_recurrence_id_only(self, coll_vdirs):
        """test searching for recurring events which only have a recuid event,
        and no master"""
        coll, vdirs = coll_vdirs
        assert len(list(coll.search('Event'))) == 0
        event = Event.fromString(
            _get_text('event_dt_recuid_no_master'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        assert len(list(coll.search('Event'))) == 1

    def test_search_recurrence_id_only_multi(self, coll_vdirs):
        """test searching for recurring events which only have a recuid event,
        and no master"""
        coll, vdirs = coll_vdirs
        assert len(list(coll.search('Event'))) == 0
        event = Event.fromString(
            _get_text('event_dt_multi_recuid_no_master'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        events = sorted(coll.search('Event'))
        assert len(events) == 2
        assert human_formatter('{start} {end} {title}')(events[0].attributes(
            dt.date.today())) == '30.06. 07:30 30.06. 12:00 Arbeit\x1b[0m'
        assert human_formatter('{start} {end} {title}')(events[1].attributes(
            dt.date.today())) == '07.07. 08:30 07.07. 12:00 Arbeit\x1b[0m'

    def test_delete_two_events(self, coll_vdirs, sleep_time):
        """testing if we can delete any of two events in two different
        calendars with the same filename"""
        coll, vdirs = coll_vdirs
        event1 = Event.fromString(
            _get_text('event_dt_simple'), calendar=cal1, locale=LOCALE_BERLIN)
        event2 = Event.fromString(
            _get_text('event_dt_simple'), calendar=cal2, locale=LOCALE_BERLIN)
        coll.insert(event1, cal1)
        sleep(sleep_time)  # make sure the etags are different
        coll.insert(event2, cal2)
        etag1 = list(vdirs[cal1].list())[0][1]
        etag2 = list(vdirs[cal2].list())[0][1]
        events = list(coll.get_localized(self.astart_berlin, self.aend_berlin))
        assert len(events) == 2
        assert events[0].calendar != events[1].calendar
        for event in events:
            if event.calendar == cal1:
                assert event.etag == etag1
            if event.calendar == cal2:
                assert event.etag == etag2

    def test_delete_recuid(self, coll_vdirs: CollVdirType):
        """Testing if we can delete a recuid (add it to exdate)"""
        coll, _ = coll_vdirs
        event_str = _get_text('event_rrule_recuid')
        event = Event.fromString(event_str, calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        event = coll.get_event('event_rrule_recurrence_id.ics', cal1)

        event = coll.delete_instance(
            'event_rrule_recurrence_id.ics',
            event.etag,
            calendar=cal1,
            rec_id=BERLIN.localize(dt.datetime(2014, 7, 14, 5)),
        )
        assert 'EXDATE;TZID=Europe/Berlin:20140714T050000' in event.raw.split()

        event = coll.delete_instance(
            'event_rrule_recurrence_id.ics',
            event.etag,
            calendar=cal1,
            rec_id=BERLIN.localize(dt.datetime(2014, 7, 21, 5)),
        )
        assert 'EXDATE;TZID=Europe/Berlin:20140714T050000,20140721T050000' in event.raw.split()

    def test_invalid_timezones(self, coll_vdirs):
        """testing if we can delete any of two events in two different
        calendars with the same filename"""
        coll, vdirs = coll_vdirs
        event = Event.fromString(
            _get_text('invalid_tzoffset'), calendar=cal1, locale=LOCALE_BERLIN)
        coll.insert(event, cal1)
        events = sorted(coll.search('Event'))
        assert len(events) == 1
        assert human_formatter('{start} {end} {title}')(events[0].attributes(dt.date.today())) == \
            '02.12. 08:00 02.12. 09:30 Some event\x1b[0m'

    def test_multi_uid_vdir(self, coll_vdirs, caplog, fix_caplog, sleep_time):
        coll, vdirs = coll_vdirs
        caplog.set_level(logging.WARNING)
        sleep(sleep_time)  # Make sure we get a new ctag on upload
        vdirs[cal1].upload(DumbItem(_get_text('event_dt_multi_uid'), uid='12345'))
        coll.update_db()
        assert list(coll.search('')) == []
        messages = [rec.message for rec in caplog.records]
        assert messages[0].startswith(
            "The .ics file at foobar/12345.ics contains multiple UIDs.\n"
        )
        assert messages[1].startswith(
            "Skipping foobar/12345.ics: \nThis event will not be available in khal."
        )


class TestDbCreation:

    def test_create_db(self, tmpdir):
        vdirpath = str(tmpdir) + '/' + cal1
        os.makedirs(vdirpath, mode=0o770)
        dbdir = str(tmpdir) + '/subdir/'
        dbpath = dbdir + 'khal.db'

        assert not os.path.isdir(dbdir)
        calendars = {cal1: {'name': cal1, 'path': vdirpath}}
        CalendarCollection(calendars, dbpath=dbpath, locale=LOCALE_BERLIN)
        assert os.path.isdir(dbdir)

    def test_failed_create_db(self, tmpdir):
        dbdir = str(tmpdir) + '/subdir/'
        dbpath = dbdir + 'khal.db'
        os.chmod(str(tmpdir), 400)

        calendars = {cal1: {'name': cal1, 'path': str(tmpdir)}}
        with pytest.raises(CouldNotCreateDbDir):
            CalendarCollection(calendars, dbpath=dbpath, locale=LOCALE_BERLIN)
        os.chmod(str(tmpdir), 777)


def test_event_different_timezones(coll_vdirs, sleep_time):
    coll, vdirs = coll_vdirs
    sleep(sleep_time)  # Make sure we get a new ctag on upload
    vdirs[cal1].upload(DumbItem(_get_text('event_dt_london'), uid='12345'))
    coll.update_db()

    events = coll.get_localized(
        BERLIN.localize(dt.datetime(2014, 4, 9, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 4, 9, 23, 59)),
    )
    events = list(events)
    assert len(events) == 1
    event = events[0]
    assert event.start_local == LONDON.localize(dt.datetime(2014, 4, 9, 14))
    assert event.end_local == LONDON.localize(dt.datetime(2014, 4, 9, 19))
    assert event.start == LONDON.localize(dt.datetime(2014, 4, 9, 14))
    assert event.end == LONDON.localize(dt.datetime(2014, 4, 9, 19))

    # no event scheduled on the next day
    events = coll.get_localized(
        BERLIN.localize(dt.datetime(2014, 4, 10, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 4, 10, 23, 59)),
    )
    events = list(events)
    assert len(events) == 0

    # now setting the local_timezone to Sydney
    coll.locale = LOCALE_SYDNEY
    events = coll.get_localized(
        SYDNEY.localize(dt.datetime(2014, 4, 9, 0, 0)),
        SYDNEY.localize(dt.datetime(2014, 4, 9, 23, 59)),
    )
    events = list(events)
    assert len(events) == 1
    event = events[0]
    assert event.start_local == SYDNEY.localize(dt.datetime(2014, 4, 9, 23))
    assert event.end_local == SYDNEY.localize(dt.datetime(2014, 4, 10, 4))
    assert event.start == LONDON.localize(dt.datetime(2014, 4, 9, 14))
    assert event.end == LONDON.localize(dt.datetime(2014, 4, 9, 19))

    # the event spans midnight Sydney, therefor it should also show up on the
    # next day
    events = coll.get_localized(SYDNEY.localize(dt.datetime(2014, 4, 10, 0, 0)),
                                SYDNEY.localize(dt.datetime(2014, 4, 10, 23, 59)))
    events = list(events)
    assert len(events) == 1
    assert event.start_local == SYDNEY.localize(dt.datetime(2014, 4, 9, 23))
    assert event.end_local == SYDNEY.localize(dt.datetime(2014, 4, 10, 4))


def test_default_calendar(coll_vdirs, sleep_time):
    """test if an update to the vdir is detected by the CalendarCollection"""
    coll, vdirs = coll_vdirs
    vdir = vdirs['foobar']
    event = coll.create_event_from_ics(event_today, 'foobar')

    assert len(list(coll.get_events_on(today))) == 0

    sleep(sleep_time)  # Make sure we get a new ctag on upload
    vdir.upload(event)
    sleep(sleep_time)
    href, etag = list(vdir.list())[0]
    assert len(list(coll.get_events_on(today))) == 0

    coll.update_db()
    sleep(sleep_time)
    assert len(list(coll.get_events_on(today))) == 1

    vdir.delete(href, etag)
    sleep(sleep_time)
    assert len(list(coll.get_events_on(today))) == 1

    coll.update_db()
    sleep(sleep_time)
    assert len(list(coll.get_events_on(today))) == 0


def test_only_update_old_event(coll_vdirs, monkeypatch, sleep_time):
    coll, vdirs = coll_vdirs

    href_one, etag_one = vdirs[cal1].upload(coll.create_event_from_ics(dedent("""
    BEGIN:VEVENT
    UID:meeting-one
    DTSTART;VALUE=DATE:20140909
    DTEND;VALUE=DATE:20140910
    SUMMARY:first meeting
    END:VEVENT
    """), cal1))

    sleep(sleep_time)  # Make sure we get a new etag for meeting-two

    href_two, etag_two = vdirs[cal1].upload(coll.create_event_from_ics(dedent("""
    BEGIN:VEVENT
    UID:meeting-two
    DTSTART;VALUE=DATE:20140910
    DTEND;VALUE=DATE:20140911
    SUMMARY:second meeting
    END:VEVENT
    """), cal1))

    sleep(sleep_time)
    coll.update_db()
    sleep(sleep_time)
    assert not coll._needs_update(cal1)

    old_update_vevent = coll._update_vevent
    updated_hrefs = []

    def _update_vevent(href, calendar):
        updated_hrefs.append(href)
        return old_update_vevent(href, calendar)
    monkeypatch.setattr(coll, '_update_vevent', _update_vevent)

    href_three, etag_three = vdirs[cal1].upload(coll.create_event_from_ics(dedent("""
    BEGIN:VEVENT
    UID:meeting-three
    DTSTART;VALUE=DATE:20140911
    DTEND;VALUE=DATE:20140912
    SUMMARY:third meeting
    END:VEVENT
    """), cal1))
    sleep(sleep_time)

    assert coll._needs_update(cal1)
    coll.update_db()
    sleep(sleep_time)
    assert updated_hrefs == [href_three]


card = """BEGIN:VCARD
VERSION:3.0
FN:Unix
BDAY:19710311
END:VCARD
"""

card_29thfeb = """BEGIN:VCARD
VERSION:3.0
FN:leapyear
BDAY:20000229
END:VCARD
"""

card_no_year = """BEGIN:VCARD
VERSION:3.0
FN:Unix
BDAY:--0311
END:VCARD
"""


def test_birthdays(coll_vdirs_birthday, sleep_time):
    coll, vdirs = coll_vdirs_birthday
    assert list(
        coll.get_floating(dt.datetime(1971, 3, 11), dt.datetime(1971, 3, 11, 23, 59, 59))
    ) == []
    sleep(sleep_time)  # Make sure we get a new ctag on upload
    vdirs[cal1].upload(DumbItem(card, 'unix'))
    coll.update_db()
    assert 'Unix\'s 41st birthday' == list(
        coll.get_floating(dt.datetime(2012, 3, 11), dt.datetime(2012, 3, 11)))[0].summary
    assert 'Unix\'s 42nd birthday' == list(
        coll.get_floating(dt.datetime(2013, 3, 11), dt.datetime(2013, 3, 11)))[0].summary
    assert 'Unix\'s 43rd birthday' == list(
        coll.get_floating(dt.datetime(2014, 3, 11), dt.datetime(2014, 3, 11)))[0].summary


def test_birthdays_29feb(coll_vdirs_birthday, sleep_time):
    """test how we deal with birthdays on 29th of feb in leap years"""
    coll, vdirs = coll_vdirs_birthday
    sleep(sleep_time)  # Make sure we get a new ctag on upload
    vdirs[cal1].upload(DumbItem(card_29thfeb, 'leap'))
    assert coll.needs_update()
    coll.update_db()
    events = list(
        coll.get_floating(dt.datetime(2004, 1, 1, 0, 0), dt.datetime(2004, 12, 31))
    )
    assert len(events) == 1
    assert events[0].summary == 'leapyear\'s 4th birthday (29th of Feb.)'
    assert events[0].start == dt.date(2004, 2, 29)
    events = list(
        coll.get_floating(dt.datetime(2005, 1, 1, 0, 0), dt.datetime(2005, 12, 31))
    )
    assert len(events) == 1
    assert events[0].summary == 'leapyear\'s 5th birthday (29th of Feb.)'
    assert events[0].start == dt.date(2005, 3, 1)
    assert list(
        coll.get_floating(dt.datetime(2001, 1, 1), dt.datetime(2001, 12, 31))
    )[0].summary == 'leapyear\'s 1st birthday (29th of Feb.)'
    assert list(
        coll.get_floating(dt.datetime(2002, 1, 1), dt.datetime(2002, 12, 31))
    )[0].summary == 'leapyear\'s 2nd birthday (29th of Feb.)'
    assert list(
        coll.get_floating(dt.datetime(2003, 1, 1), dt.datetime(2003, 12, 31))
    )[0].summary == 'leapyear\'s 3rd birthday (29th of Feb.)'
    assert list(
        coll.get_floating(dt.datetime(2023, 1, 1), dt.datetime(2023, 12, 31))
    )[0].summary == 'leapyear\'s 23rd birthday (29th of Feb.)'
    assert events[0].start == dt.date(2005, 3, 1)


def test_birthdays_no_year(coll_vdirs_birthday, sleep_time):
    coll, vdirs = coll_vdirs_birthday
    assert list(
        coll.get_floating(dt.datetime(1971, 3, 11), dt.datetime(1971, 3, 11, 23, 59, 59))
    ) == []
    sleep(sleep_time)  # Make sure we get a new ctag on upload
    vdirs[cal1].upload(DumbItem(card_no_year, 'vcard.vcf'))
    coll.update_db()
    events = list(coll.get_floating(dt.datetime(1971, 3, 11), dt.datetime(1971, 3, 11, 23, 59, 59)))
    assert len(events) == 1
    assert 'Unix\'s birthday' == events[0].summary



import datetime as dt
import random
import textwrap

import icalendar
from freezegun import freeze_time

from khal.icalendar import new_vevent, split_ics

from .utils import LOCALE_BERLIN, _get_text, _replace_uid, normalize_component


def _get_TZIDs(lines):
    """from a list of strings, get all unique strings that start with TZID"""
    return sorted(line for line in lines if line.startswith('TZID'))


def test_normalize_component():
    assert normalize_component(textwrap.dedent("""
    BEGIN:VEVENT
    DTSTART;TZID=Europe/Berlin:20140409T093000
    END:VEVENT
    """)) != normalize_component(textwrap.dedent("""
    BEGIN:VEVENT
    DTSTART;TZID=Oyrope/Berlin:20140409T093000
    END:VEVENT
    """))


def test_new_vevent():
    with freeze_time('20220702T1400'):
        vevent = _replace_uid(new_vevent(
            LOCALE_BERLIN,
            dt.date(2022, 7, 2),
            dt.date(2022, 7, 3),
            'An Event',
            allday=True,
            repeat='weekly',
        ))
        assert vevent.to_ical().decode('utf-8') == '\r\n'.join([
            'BEGIN:VEVENT',
            'SUMMARY:An Event',
            'DTSTART;VALUE=DATE:20220702',
            'DTEND;VALUE=DATE:20220703',
            'DTSTAMP:20220702T140000Z',
            'UID:E41JRQX2DB4P1AQZI86BAT7NHPBHPRIIHQKA',
            'RRULE:FREQ=WEEKLY',
            'END:VEVENT',
            ''
        ])


def test_split_ics():
    cal = _get_text('cal_lots_of_timezones')
    vevents = split_ics(cal)

    vevents0 = vevents[0].split('\r\n')
    vevents1 = vevents[1].split('\r\n')

    part0 = _get_text('part0').split('\n')
    part1 = _get_text('part1').split('\n')

    assert _get_TZIDs(vevents0) == _get_TZIDs(part0)
    assert _get_TZIDs(vevents1) == _get_TZIDs(part1)

    assert sorted(vevents0) == sorted(part0)
    assert sorted(vevents1) == sorted(part1)


def test_split_ics_random_uid():
    random.seed(123)
    cal = _get_text('cal_lots_of_timezones')
    vevents = split_ics(cal, random_uid=True)

    part0 = _get_text('part0').split('\n')
    part1 = _get_text('part1').split('\n')

    for item in icalendar.Calendar.from_ical(vevents[0]).walk():
        if item.name == 'VEVENT':
            assert item['UID'] == 'DRF0RGCY89VVDKIV9VPKA1FYEAU2GCFJIBS1'
    for item in icalendar.Calendar.from_ical(vevents[1]).walk():
        if item.name == 'VEVENT':
            assert item['UID'] == '4Q4CTV74N7UAZ618570X6CLF5QKVV9ZE3YVB'

    # after replacing the UIDs, everything should be as above
    vevents0 = vevents[0].replace('DRF0RGCY89VVDKIV9VPKA1FYEAU2GCFJIBS1', '123').split('\r\n')
    vevents1 = vevents[1].replace('4Q4CTV74N7UAZ618570X6CLF5QKVV9ZE3YVB', 'abcde').split('\r\n')

    assert _get_TZIDs(vevents0) == _get_TZIDs(part0)
    assert _get_TZIDs(vevents1) == _get_TZIDs(part1)

    assert sorted(vevents0) == sorted(part0)
    assert sorted(vevents1) == sorted(part1)


def test_split_ics_missing_timezone():
    """testing if we detect the missing timezone in splitting"""
    cal = _get_text('event_dt_local_missing_tz')
    split_ics(cal, random_uid=True, default_timezone=LOCALE_BERLIN['default_timezone'])


def test_windows_timezone(caplog):
    """Test if a windows tz format works"""
    cal = _get_text("tz_windows_format")
    split_ics(cal)
    assert "Cannot find timezone `Pacific/Auckland`" not in caplog.text


def test_split_ics_without_uid():
    cal = _get_text('without_uid')
    vevents = split_ics(cal)
    assert vevents
    vevents2 = split_ics(cal)
    assert vevents[0] == vevents2[0]



import os
from typing import Dict, Tuple

import icalendar
import pytz

from khal.custom_types import LocaleConfiguration
from khal.khalendar import CalendarCollection
from khal.khalendar.vdir import Vdir

CollVdirType = Tuple[CalendarCollection, Dict[str, Vdir]]

cal0 = 'a_calendar'
cal1 = 'foobar'
cal2 = "Dad's calendar"
cal3 = 'private'

example_cals = [cal0, cal1, cal2, cal3]

BERLIN = pytz.timezone('Europe/Berlin')
NEW_YORK = pytz.timezone('America/New_York')
LONDON = pytz.timezone('Europe/London')
SAMOA = pytz.timezone('Pacific/Samoa')
SYDNEY = pytz.timezone('Australia/Sydney')
GMTPLUS3 = pytz.timezone('Etc/GMT+3')
# the lucky people in Bogota don't know the pain that is DST
BOGOTA = pytz.timezone('America/Bogota')


LOCALE_BERLIN: LocaleConfiguration = {
    'default_timezone': BERLIN,
    'local_timezone': BERLIN,
    'dateformat': '%d.%m.',
    'longdateformat': '%d.%m.%Y',
    'timeformat': '%H:%M',
    'datetimeformat': '%d.%m. %H:%M',
    'longdatetimeformat': '%d.%m.%Y %H:%M',
    'unicode_symbols': True,
    'firstweekday': 0,
    'weeknumbers': False,
}

LOCALE_NEW_YORK: LocaleConfiguration = {
    'default_timezone': NEW_YORK,
    'local_timezone': NEW_YORK,
    'timeformat': '%H:%M',
    'dateformat': '%Y/%m/%d',
    'longdateformat': '%Y/%m/%d',
    'datetimeformat': '%Y/%m/%d-%H:%M',
    'longdatetimeformat': '%Y/%m/%d-%H:%M',
    'firstweekday': 6,
    'unicode_symbols': True,
    'weeknumbers': False,
}

LOCALE_SAMOA = {
    'local_timezone': SAMOA,
    'default_timezone': SAMOA,
    'unicode_symbols': True,
}
LOCALE_SYDNEY = {'local_timezone': SYDNEY, 'default_timezone': SYDNEY}

LOCALE_BOGOTA = LOCALE_BERLIN.copy()
LOCALE_BOGOTA['local_timezone'] = BOGOTA
LOCALE_BOGOTA['default_timezone'] = BOGOTA

LOCALE_MIXED = LOCALE_BERLIN.copy()
LOCALE_MIXED['local_timezone'] = BOGOTA

LOCALE_FLOATING = LOCALE_BERLIN.copy()
LOCALE_FLOATING['default_timezone'] = None  # type: ignore
LOCALE_FLOATING['local_timezone'] = None  # type: ignore


def normalize_component(x):
    x = icalendar.cal.Component.from_ical(x)

    def inner(c):
        contentlines = icalendar.cal.Contentlines()
        for name, value in c.property_items(sorted=True, recursive=False):
            contentlines.append(c.content_line(name, value, sorted=True))
        contentlines.append('')

        return (c.name, contentlines.to_ical(),
                frozenset(inner(sub) for sub in c.subcomponents))

    return inner(x)


def _get_text(event_name):
    directory = '/'.join(__file__.split('/')[:-1]) + '/ics/'
    if directory == '/ics/':
        directory = './ics/'

    with open(os.path.join(directory, event_name + '.ics'), 'rb') as f:
        rv = f.read().decode('utf-8')

    return rv


def _get_vevent_file(event_path):
    directory = '/'.join(__file__.split('/')[:-1]) + '/ics/'
    with open(os.path.join(directory, event_path + '.ics'), 'rb') as f:
        ical = icalendar.Calendar.from_ical(
            f.read()
        )
    for component in ical.walk():
        if component.name == 'VEVENT':
            return component


def _get_ics_filepath(event_name):
    directory = '/'.join(__file__.split('/')[:-1]) + '/ics/'
    if directory == '/ics/':
        directory = './ics/'
    return os.path.join(directory, event_name + '.ics')


def _get_all_vevents_file(event_path):
    directory = '/'.join(__file__.split('/')[:-1]) + '/ics/'
    ical = icalendar.Calendar.from_ical(
        open(os.path.join(directory, event_path + '.ics'), 'rb').read()
    )
    for component in ical.walk():
        if component.name == 'VEVENT':
            yield component


def _replace_uid(event):
    """
    Replace an event's UID with E41JRQX2DB4P1AQZI86BAT7NHPBHPRIIHQKA.
    """
    event.pop('uid')
    event.add('uid', 'E41JRQX2DB4P1AQZI86BAT7NHPBHPRIIHQKA')
    return event


class DumbItem:
    def __init__(self, raw, uid) -> None:
        self.raw = raw
        self.uid = uid



from khal.ui.widgets import delete_last_word


def test_delete_last_word():

    tests = [
        ('Fü1ü  Bär!', 'Fü1ü  Bär', 1),
        ('Füü Bär1', 'Füü ', 1),
        ('Füü1 Bär1', 'Füü1 ', 1),
        (' Füü Bär', ' Füü ', 1),
        ('Füü Bär.Füü', 'Füü Bär.', 1),
        ('Füü Bär.(Füü)', 'Füü Bär.(Füü', 1),
        ('Füü ', '', 1),
        ('Füü  ', '', 1),
        ('Füü', '', 1),
        ('', '', 1),

        ('Füü Bär.(Füü)', 'Füü Bär.', 3),
        ('Füü Bär1', '', 2),
        ('Lorem ipsum dolor sit amet, consetetur sadipscing elitr, '
         'sed diam nonumy eirmod tempor invidunt ut labore et dolore '
         'magna aliquyam erat, sed diam volest.',
         'Lorem ipsum dolor sit amet, consetetur sadipscing elitr, '
         'sed diam nonumy eirmod tempor invidunt ut labore ',
         10)
    ]

    for org, short, number in tests:
        assert delete_last_word(org, number) == short



from khal.terminal import colored, merge_columns


def test_colored():
    assert colored('test', 'light cyan') == '\33[1;36mtest\x1b[0m'
    assert colored('täst', 'white') == '\33[37mtäst\x1b[0m'
    assert colored('täst', 'white', 'dark green') == '\x1b[37m\x1b[42mtäst\x1b[0m'
    assert colored('täst', 'light magenta', 'dark green', True) == '\x1b[1;35m\x1b[42mtäst\x1b[0m'
    assert colored('täst', 'light magenta', 'dark green', False) == '\x1b[95m\x1b[42mtäst\x1b[0m'
    assert colored('täst', 'light magenta', 'light green', True) == '\x1b[1;35m\x1b[42mtäst\x1b[0m'
    assert colored('täst', 'light magenta', 'light green', False) == '\x1b[95m\x1b[102mtäst\x1b[0m'
    assert colored('täst', '5', '20') == '\x1b[38;5;5m\x1b[48;5;20mtäst\x1b[0m'
    assert colored('täst', '#F0F', '#00AABB') == \
        '\x1b[38;2;255;0;255m\x1b[48;2;0;170;187mtäst\x1b[0m'


class TestMergeColumns:

    def test_longer_right(self):
        left = ['uiae', 'nrtd']
        right = ['123456', '234567', '345678']
        out = ['uiae    123456',
               'nrtd    234567',
               '        345678']
        assert merge_columns(left, right, width=4) == out

    def test_longer_left(self):
        left = ['uiae', 'nrtd', 'xvlc']
        right = ['123456', '234567']
        out = ['uiae    123456', 'nrtd    234567', 'xvlc    ']
        assert merge_columns(left, right, width=4) == out



import logging
import os
from time import sleep

import pytest

from khal.custom_types import CalendarConfiguration
from khal.khalendar import CalendarCollection
from khal.khalendar.vdir import Vdir

from .utils import LOCALE_BERLIN, CollVdirType, cal1, example_cals


@pytest.fixture
def metavdirs(tmpdir):
    tmpdir = str(tmpdir)
    dirstructure = [
        '/cal1/public/',
        '/cal1/private/',
        '/cal2/public/',
        '/cal3/public/',
        '/cal3/work/',
        '/cal3/home/',
        '/cal4/cfgcolor/',
        '/cal4/dircolor/',
        '/cal4/cfgcolor_again/',
        '/cal4/cfgcolor_once_more/',
        '/singlecollection/',
    ]
    for one in dirstructure:
        os.makedirs(tmpdir + one)
    filestructure = [
        ('/cal1/public/displayname', 'my calendar'),
        ('/cal1/public/color', 'dark blue'),
        ('/cal1/private/displayname', 'my private calendar'),
        ('/cal1/private/color', '#FF00FF'),
        ('/cal4/dircolor/color', 'dark blue'),
    ]
    for filename, content in filestructure:
        with open(tmpdir + filename, 'w') as metafile:
            metafile.write(content)
    return tmpdir


@pytest.fixture
def coll_vdirs(tmpdir) -> CollVdirType:
    calendars, vdirs = {}, {}
    for name in example_cals:
        path = str(tmpdir) + '/' + name
        os.makedirs(path, mode=0o770)
        readonly = True if name == 'a_calendar' else False
        calendars[name] = CalendarConfiguration(
            name=name,
            path=path,
            readonly=readonly,
            color='dark blue',
            priority=10,
            ctype='calendar',
            addresses='user@example.com',
        )
        vdirs[name] = Vdir(path, '.ics')
    coll = CalendarCollection(calendars=calendars, dbpath=':memory:', locale=LOCALE_BERLIN)
    coll.default_calendar_name = cal1
    return coll, vdirs


@pytest.fixture
def coll_vdirs_birthday(tmpdir):
    calendars, vdirs = {}, {}
    for name in example_cals:
        path = str(tmpdir) + '/' + name
        os.makedirs(path, mode=0o770)
        readonly = True if name == 'a_calendar' else False
        calendars[name] = {'name': name, 'path': path, 'color': 'dark blue',
                           'readonly': readonly, 'unicode_symbols': True, 'ctype': 'birthdays',
                           'addresses': 'user@example.com'}
        vdirs[name] = Vdir(path, '.vcf')
    coll = CalendarCollection(calendars=calendars, dbpath=':memory:', locale=LOCALE_BERLIN)
    coll.default_calendar_name = cal1
    return coll, vdirs


@pytest.fixture(autouse=True)
def never_echo_bytes(monkeypatch):
    '''Click's echo function will not strip colorcodes if we call `click.echo`
    with a bytestring message. The reason for this that bytestrings may contain
    arbitrary binary data (such as images).

    Khal is not concerned with such data at all, but may contain a few
    instances where it explicitly encodes its output into the configured
    locale. This in turn would break the functionality of the global
    `--color/--no-color` flag.
    '''
    from click import echo as old_echo

    def echo(msg=None, *a, **kw):
        assert not isinstance(msg, bytes)
        return old_echo(msg, *a, **kw)

    monkeypatch.setattr('click.echo', echo)

    class Result:
        @staticmethod
        def undo():
            monkeypatch.setattr('click.echo', old_echo)

    return Result


@pytest.fixture(scope='session')
def sleep_time(tmpdir_factory):
    """
    Returns the filesystem's mtime precision

    Returns how long we need to sleep for the filesystem's mtime precision to
    pick up differences.

    This keeps test fast on systems with high precisions, but makes them pass
    on those that don't.
    """
    tmpfile = tmpdir_factory.mktemp('sleep').join('touch_me')

    def touch_and_mtime():
        tmpfile.open('w').close()
        stat = os.stat(str(tmpfile))
        return getattr(stat, 'st_mtime_ns', stat.st_mtime)

    i = 0.00001
    while i < 100:
        # Measure three times to avoid things like 12::18:11.9994 [mis]passing
        first = touch_and_mtime()
        sleep(i)
        second = touch_and_mtime()
        sleep(i)
        third = touch_and_mtime()

        if first != second != third:
            return i * 1.1
        i = i * 10

    # This should never happen, but oh, well:
    raise Exception(
        'Filesystem does not seem to save modified times of files. \n'
        'Cannot run tests that depend on this.'
    )


@pytest.fixture
def fix_caplog(monkeypatch):
    """Temporarily undoes the logging setup by click-log such that the caplog fixture can be used"""
    logger = logging.getLogger('khal')
    monkeypatch.setattr(logger, 'handlers', [])
    monkeypatch.setattr(logger, 'propagate', True)



import click
import pytest

from khal.configwizard import get_collection_names_from_vdirs, validate_int


def test_validate_int():
    assert validate_int('3', 0, 3) == 3
    with pytest.raises(click.UsageError):
        validate_int('3', 0, 2)
    with pytest.raises(click.UsageError):
        validate_int('two', 0, 2)


def test_default_vdir(metavdirs):
    names = get_collection_names_from_vdirs([('found', f'{metavdirs}/**/', 'discover')])
    assert names == [
        'my private calendar', 'my calendar', 'public', 'home', 'public1', 'work',
        'cfgcolor', 'cfgcolor_again', 'cfgcolor_once_more', 'dircolor', 'singlecollection',
    ]



import datetime as dt

from freezegun import freeze_time

from khal.ui import DayWalker, DListBox, StaticDayWalker

from ..utils import LOCALE_BERLIN
from .canvas_render import CanvasTranslator

CONF = {'locale': LOCALE_BERLIN, 'keybindings': {},
        'view': {'monthdisplay': 'firstday'},
        'default': {'timedelta': dt.timedelta(days=3)},
        }


palette = {
    'date header focused': 'blue',
    'date header': 'green',
    'default': 'black',
}


@freeze_time('2017-6-7')
def test_daywalker(coll_vdirs):
    collection, _ = coll_vdirs
    this_date = dt.date.today()
    daywalker = DayWalker(this_date, None, CONF, collection, delete_status={})
    elistbox = DListBox(
        daywalker, parent=None, conf=CONF,
        delete_status=lambda: False,
        toggle_delete_all=None,
        toggle_delete_instance=None,
        dynamic_days=True,
    )
    canvas = elistbox.render((50, 6), True)
    assert CanvasTranslator(canvas, palette).transform() == \
        """\x1b[34mToday (Wednesday, 07.06.2017)\x1b[0m
\x1b[32mTomorrow (Thursday, 08.06.2017)\x1b[0m
\x1b[32mFriday, 09.06.2017 (2 days from now)\x1b[0m
\x1b[32mSaturday, 10.06.2017 (3 days from now)\x1b[0m
\x1b[32mSunday, 11.06.2017 (4 days from now)\x1b[0m
\x1b[32mMonday, 12.06.2017 (5 days from now)\x1b[0m
"""


@freeze_time('2017-6-7')
def test_staticdaywalker(coll_vdirs):
    collection, _ = coll_vdirs
    this_date = dt.date.today()
    daywalker = StaticDayWalker(this_date, None, CONF, collection, delete_status={})
    elistbox = DListBox(
        daywalker, parent=None, conf=CONF,
        delete_status=lambda: False,
        toggle_delete_all=None,
        toggle_delete_instance=None,
        dynamic_days=False,
    )
    canvas = elistbox.render((50, 10), True)
    assert CanvasTranslator(canvas, palette).transform() == \
        """\x1b[34mToday (Wednesday, 07.06.2017)\x1b[0m
\x1b[32mTomorrow (Thursday, 08.06.2017)\x1b[0m
\x1b[32mFriday, 09.06.2017 (2 days from now)\x1b[0m







"""


@freeze_time('2017-6-7')
def test_staticdaywalker_3(coll_vdirs):
    collection, _ = coll_vdirs
    this_date = dt.date.today()
    conf = {}
    conf.update(CONF)
    conf['default'] = {'timedelta': dt.timedelta(days=1)}
    daywalker = StaticDayWalker(this_date, None, conf, collection, delete_status={})
    elistbox = DListBox(
        daywalker, parent=None, conf=conf,
        delete_status=lambda: False,
        toggle_delete_all=None,
        toggle_delete_instance=None,
        dynamic_days=False,
    )
    canvas = elistbox.render((50, 10), True)
    assert CanvasTranslator(canvas, palette).transform() == \
        '\x1b[34mToday (Wednesday, 07.06.2017)\x1b[0m\n\n\n\n\n\n\n\n\n\n'



import datetime as dt
from collections import OrderedDict

import pytest
from freezegun import freeze_time

from khal.exceptions import DateTimeParseError, FatalError
from khal.icalendar import new_vevent
from khal.parse_datetime import (
    construct_daynames,
    eventinfofstr,
    guessdatetimefstr,
    guessrangefstr,
    guesstimedeltafstr,
    timedelta2str,
    weekdaypstr,
)

from .utils import (
    LOCALE_BERLIN,
    LOCALE_FLOATING,
    LOCALE_NEW_YORK,
    _replace_uid,
    normalize_component,
)


def _create_testcases(*cases):
    return [(userinput, ('\r\n'.join(output) + '\r\n').encode('utf-8'))
            for userinput, output in cases]


def _construct_event(info, locale,
                     defaulttimelen=60, defaultdatelen=1, description=None,
                     location=None, categories=None, repeat=None, until=None,
                     alarm=None, **kwargs):
    info = eventinfofstr(' '.join(info), locale,
                         default_event_duration=dt.timedelta(hours=1),
                         default_dayevent_duration=dt.timedelta(days=1),
                         adjust_reasonably=True,
                         )
    if description is not None:
        info["description"] = description
    event = new_vevent(
        locale=locale, location=location,
        categories=categories, repeat=repeat, until=until,
        alarms=alarm,
        **info)
    return event


def _create_vevent(*args):
    """
    Adapt and return a default vevent for testing.

    Accepts an arbitrary amount of strings like 'DTSTART;VALUE=DATE:2013015'.
    Updates the default vevent if the key (the first word) is found and
    appends the value otherwise.
    """
    def_vevent = OrderedDict(
        [('BEGIN', 'BEGIN:VEVENT'),
         ('SUMMARY', 'SUMMARY:Äwesöme Event'),
         ('DTSTART', 'DTSTART;VALUE=DATE:20131025'),
         ('DTEND', 'DTEND;VALUE=DATE:20131026'),
         ('DTSTAMP', 'DTSTAMP:20140216T120000Z'),
         ('UID', 'UID:E41JRQX2DB4P1AQZI86BAT7NHPBHPRIIHQKA')])

    for row in args:
        key = row.replace(':', ';').split(';')[0]
        def_vevent[key] = row

    def_vevent['END'] = 'END:VEVENT'
    return list(def_vevent.values())


class TestTimeDelta2Str:

    def test_single(self):
        assert timedelta2str(dt.timedelta(minutes=10)) == '10m'

    def test_negative(self):
        assert timedelta2str(dt.timedelta(minutes=-10)) == '-10m'

    def test_days(self):
        assert timedelta2str(dt.timedelta(days=2)) == '2d'

    def test_multi(self):
        assert timedelta2str(
            dt.timedelta(days=6, hours=-3, minutes=10, seconds=-3)
        ) == '5d 21h 9m 57s'


def test_weekdaypstr():
    for string, weekdayno in [
            ('monday', 0),
            ('tue', 1),
            ('wednesday', 2),
            ('thursday', 3),
            ('fri', 4),
            ('saturday', 5),
            ('sun', 6),
    ]:
        assert weekdaypstr(string) == weekdayno


def test_weekdaypstr_invalid():
    with pytest.raises(ValueError):
        weekdaypstr('foobar')


def test_construct_daynames():
    with freeze_time('2016-9-19'):
        assert construct_daynames(dt.date(2016, 9, 19)) == 'Today'
        assert construct_daynames(dt.date(2016, 9, 20)) == 'Tomorrow'
        assert construct_daynames(dt.date(2016, 9, 21)) == 'Wednesday'


class TestGuessDatetimefstr:

    @freeze_time('2016-9-19T8:00')
    def test_today(self):
        assert (dt.datetime(2016, 9, 19, 13), False) == \
            guessdatetimefstr(['today', '13:00'], LOCALE_BERLIN)
        assert dt.date.today() == guessdatetimefstr(['today'], LOCALE_BERLIN)[0].date()

    @freeze_time('2016-9-19T8:00')
    def test_tomorrow(self):
        assert (dt.datetime(2016, 9, 20, 16), False) == \
            guessdatetimefstr('tomorrow 16:00 16:00'.split(), locale=LOCALE_BERLIN)

    @freeze_time('2016-9-19T8:00')
    def test_time_tomorrow(self):
        assert (dt.datetime(2016, 9, 20, 16), False) == \
            guessdatetimefstr(
                '16:00'.split(), locale=LOCALE_BERLIN, default_day=dt.date(2016, 9, 20))

    @freeze_time('2016-9-19T8:00')
    def test_time_yesterday(self):
        assert (dt.datetime(2016, 9, 18, 16), False) == guessdatetimefstr(
            'Yesterday 16:00'.split(),
            locale=LOCALE_BERLIN,
            default_day=dt.datetime.today())

    @freeze_time('2016-9-19')
    def test_time_weekday(self):
        assert (dt.datetime(2016, 9, 23, 16), False) == guessdatetimefstr(
            'Friday 16:00'.split(),
            locale=LOCALE_BERLIN,
            default_day=dt.datetime.today())

    @freeze_time('2016-9-19 17:53')
    def test_time_now(self):
        assert (dt.datetime(2016, 9, 19, 17, 53), False) == guessdatetimefstr(
            'now'.split(), locale=LOCALE_BERLIN, default_day=dt.datetime.today())

    @freeze_time('2016-12-30 17:53')
    def test_long_not_configured(self):
        """long version is not configured, but short contains the year"""
        locale = {
            'timeformat': '%H:%M',
            'dateformat': '%Y-%m-%d',
            'longdateformat': '',
            'datetimeformat': '%Y-%m-%d %H:%M',
            'longdatetimeformat': '',
        }
        assert (dt.datetime(2017, 1, 1), True) == guessdatetimefstr(
            '2017-1-1'.split(), locale=locale, default_day=dt.datetime.today())
        assert (dt.datetime(2017, 1, 1, 16, 30), False) == guessdatetimefstr(
            '2017-1-1 16:30'.split(), locale=locale, default_day=dt.datetime.today())

    @freeze_time('2016-12-30 17:53')
    def test_short_format_contains_year(self):
        """if the non long versions of date(time)format contained a year, the
        current year would be used instead of the given one, see #545"""
        locale = {
            'timeformat': '%H:%M',
            'dateformat': '%Y-%m-%d',
            'longdateformat': '%Y-%m-%d',
            'datetimeformat': '%Y-%m-%d %H:%M',
            'longdatetimeformat': '%Y-%m-%d %H:%M',
        }
        assert (dt.datetime(2017, 1, 1), True) == guessdatetimefstr(
            '2017-1-1'.split(), locale=locale, default_day=dt.datetime.today())
        assert (dt.datetime(2017, 1, 1, 16, 30), False) == guessdatetimefstr(
            '2017-1-1 16:30'.split(), locale=locale, default_day=dt.datetime.today())


class TestGuessTimedeltafstr:

    def test_single(self):
        assert dt.timedelta(minutes=10) == guesstimedeltafstr('10m')

    def test_seconds(self):
        assert dt.timedelta(seconds=10) == guesstimedeltafstr('10s')

    def test_single_plus(self):
        assert dt.timedelta(minutes=10) == guesstimedeltafstr('+10m')

    def test_seconds_plus(self):
        assert dt.timedelta(seconds=10) == guesstimedeltafstr('+10s')

    def test_days_plus(self):
        assert dt.timedelta(days=10) == guesstimedeltafstr('+10days')

    def test_negative(self):
        assert dt.timedelta(minutes=-10) == guesstimedeltafstr('-10m')

    def test_multi(self):
        assert dt.timedelta(days=1, hours=-3, minutes=10) == \
            guesstimedeltafstr(' 1d -3H 10min ')

    def test_multi_plus(self):
        assert dt.timedelta(days=1, hours=3, minutes=10) == \
            guesstimedeltafstr(' 1d +3H 10min ')

    def test_multi_plus_minus(self):
        assert dt.timedelta(days=0, hours=21, minutes=10) == \
            guesstimedeltafstr('+1d -3H 10min ')

    def test_multi_nospace(self):
        assert dt.timedelta(days=1, hours=-3, minutes=10) == \
            guesstimedeltafstr('1D-3hour10m')

    def test_garbage(self):
        with pytest.raises(ValueError):
            guesstimedeltafstr('10mbar')

    def test_moregarbage(self):
        with pytest.raises(ValueError):
            guesstimedeltafstr('foo10m')

    def test_same(self):
        assert dt.timedelta(minutes=20) == \
            guesstimedeltafstr('10min 10minutes')


class TestGuessRangefstr:

    @freeze_time('2016-9-19')
    def test_today(self):
        assert (dt.datetime(2016, 9, 19, 13), dt.datetime(2016, 9, 19, 14), False) == \
            guessrangefstr('13:00 14:00', locale=LOCALE_BERLIN)
        assert (dt.datetime(2016, 9, 19), dt.datetime(2016, 9, 21), True) == \
            guessrangefstr('today tomorrow', LOCALE_BERLIN)

    @freeze_time('2016-9-19 16:34')
    def test_tomorrow(self):
        # XXX remove this funtionality, we shouldn't support this anyway
        assert (dt.datetime(2016, 9, 19), dt.datetime(2016, 9, 21, 16), True) == \
            guessrangefstr('today tomorrow 16:00', locale=LOCALE_BERLIN)

    @freeze_time('2016-9-19 13:34')
    def test_time_tomorrow(self):
        assert (dt.datetime(2016, 9, 19, 16), dt.datetime(2016, 9, 19, 17), False) == \
            guessrangefstr('16:00', locale=LOCALE_BERLIN)
        assert (dt.datetime(2016, 9, 19, 16), dt.datetime(2016, 9, 19, 17), False) == \
            guessrangefstr('16:00 17:00', locale=LOCALE_BERLIN)

    def test_start_and_end_date(self):
        assert (dt.datetime(2016, 1, 1), dt.datetime(2017, 1, 2), True) == \
            guessrangefstr('1.1.2016 1.1.2017', locale=LOCALE_BERLIN)

    def test_start_and_no_end_date(self):
        assert (dt.datetime(2016, 1, 1), dt.datetime(2016, 1, 2), True) == \
            guessrangefstr('1.1.2016', locale=LOCALE_BERLIN)

    def test_start_and_end_date_time(self):
        assert (dt.datetime(2016, 1, 1, 10), dt.datetime(2017, 1, 1, 22), False) == \
            guessrangefstr(
                '1.1.2016 10:00 1.1.2017 22:00', locale=LOCALE_BERLIN)

    def test_start_and_eod(self):
        start, end = dt.datetime(2016, 1, 1, 10), dt.datetime(2016, 1, 1, 23, 59, 59, 999999)
        assert (start, end, False) == guessrangefstr('1.1.2016 10:00 eod', locale=LOCALE_BERLIN)

    def test_start_and_week(self):
        assert (dt.datetime(2015, 12, 28), dt.datetime(2016, 1, 5), True) == \
            guessrangefstr('1.1.2016 week', locale=LOCALE_BERLIN)

    def test_start_and_delta_1d(self):
        assert (dt.datetime(2016, 1, 1), dt.datetime(2016, 1, 2), True) == \
            guessrangefstr('1.1.2016 1d', locale=LOCALE_BERLIN)

    def test_start_and_delta_3d(self):
        assert (dt.datetime(2016, 1, 1), dt.datetime(2016, 1, 4), True) == \
            guessrangefstr('1.1.2016 3d', locale=LOCALE_BERLIN)

    def test_start_dt_and_delta(self):
        assert (dt.datetime(2016, 1, 1, 10), dt.datetime(2016, 1, 4, 10), False) == \
            guessrangefstr('1.1.2016 10:00 3d', locale=LOCALE_BERLIN)

    def test_start_allday_and_delta_datetime(self):
        with pytest.raises(FatalError):
            guessrangefstr('1.1.2016 3d3m', locale=LOCALE_BERLIN)

    def test_start_zero_day_delta(self):
        with pytest.raises(FatalError):
            guessrangefstr('1.1.2016 0d', locale=LOCALE_BERLIN)

    @freeze_time('20160216')
    def test_week(self):
        assert (dt.datetime(2016, 2, 15), dt.datetime(2016, 2, 23), True) == \
            guessrangefstr('week', locale=LOCALE_BERLIN)

    def test_invalid(self):
        with pytest.raises(DateTimeParseError):
            guessrangefstr('3d', locale=LOCALE_BERLIN)
        with pytest.raises(DateTimeParseError):
            guessrangefstr('35.1.2016', locale=LOCALE_BERLIN)
        with pytest.raises(DateTimeParseError):
            guessrangefstr('1.1.2016 2x', locale=LOCALE_BERLIN)
        with pytest.raises(DateTimeParseError):
            guessrangefstr('1.1.2016x', locale=LOCALE_BERLIN)
        with pytest.raises(DateTimeParseError):
            guessrangefstr('xxx yyy zzz', locale=LOCALE_BERLIN)

    @freeze_time('2016-12-30 17:53')
    def test_short_format_contains_year(self):
        """if the non long versions of date(time)format contained a year, the
        current year would be used instead of the given one, see #545

        same as above, but for guessrangefstr
        """
        locale = {
            'timeformat': '%H:%M',
            'dateformat': '%Y-%m-%d',
            'longdateformat': '%Y-%m-%d',
            'datetimeformat': '%Y-%m-%d %H:%M',
            'longdatetimeformat': '%Y-%m-%d %H:%M',
        }
        assert (dt.datetime(2017, 1, 1), dt.datetime(2017, 1, 2), True) == \
            guessrangefstr('2017-1-1 2017-1-1', locale=locale)


test_set_format_de = _create_testcases(
    # all-day-events
    # one day only
    ('25.10.2013 Äwesöme Event',
     _create_vevent('DTSTART;VALUE=DATE:20131025',
                    'DTEND;VALUE=DATE:20131026')),

    # 2 day
    ('15.08.2014 16.08. Äwesöme Event',
     _create_vevent('DTSTART;VALUE=DATE:20140815',
                    'DTEND;VALUE=DATE:20140817')),  # XXX

    # end date in next year and not specified
    ('29.12.2014 03.01. Äwesöme Event',
     _create_vevent('DTSTART;VALUE=DATE:20141229',
                    'DTEND;VALUE=DATE:20150104')),

    # end date in next year
    ('29.12.2014 03.01.2015 Äwesöme Event',
     _create_vevent('DTSTART;VALUE=DATE:20141229',
                    'DTEND;VALUE=DATE:20150104')),

    # datetime events
    # start and end date same, no explicit end date given
    ('25.10.2013 18:00 20:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20131025T180000',
         'DTEND;TZID=Europe/Berlin:20131025T200000')),

    # start and end date same, ends 24:00 which should be 00:00 (start) of next
    # day
    ('25.10.2013 18:00 24:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20131025T180000',
         'DTEND;TZID=Europe/Berlin:20131026T000000')),

    # start and end date same, explicit end date (but no year) given
    ('25.10.2013 18:00 26.10. 20:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20131025T180000',
         'DTEND;TZID=Europe/Berlin:20131026T200000')),

    ('30.12.2013 18:00 2.1. 20:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20131230T180000',
         'DTEND;TZID=Europe/Berlin:20140102T200000')),

    # only start date given (no year, past day and month)
    ('25.01. 18:00 20:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20150125T180000',
         'DTEND;TZID=Europe/Berlin:20150125T200000')),

    # date ends next day, but end date not given
    ('25.10.2013 23:00 0:30 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20131025T230000',
         'DTEND;TZID=Europe/Berlin:20131026T003000')),

    ('2.2. 23:00 0:30 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20150202T230000',
         'DTEND;TZID=Europe/Berlin:20150203T003000')),

    # only start datetime given
    ('25.10.2013 06:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20131025T060000',
         'DTEND;TZID=Europe/Berlin:20131025T070000')),

    # timezone given
    ('25.10.2013 06:00 America/New_York Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=America/New_York:20131025T060000',
         'DTEND;TZID=America/New_York:20131025T070000'))
)


@freeze_time('20140216T120000')
def test_construct_event_format_de():
    for data_list, vevent_expected in test_set_format_de:
        vevent = _construct_event(data_list.split(), locale=LOCALE_BERLIN)
        assert _replace_uid(vevent).to_ical() == vevent_expected


test_set_format_us = _create_testcases(
    ('1999/12/31-06:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=America/New_York:19991231T060000',
         'DTEND;TZID=America/New_York:19991231T070000')),

    ('2014/12/18 2014/12/20 Äwesöme Event',
     _create_vevent('DTSTART;VALUE=DATE:20141218',
                    'DTEND;VALUE=DATE:20141221')),
)


@freeze_time('2014-02-16 12:00:00')
def test__construct_event_format_us():
    for data_list, vevent in test_set_format_us:
        event = _construct_event(data_list.split(), locale=LOCALE_NEW_YORK)
        assert _replace_uid(event).to_ical() == vevent


test_set_format_de_complexer = _create_testcases(
    # now events where the start date has to be inferred, too
    # today
    ('8:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20140216T080000',
         'DTEND;TZID=Europe/Berlin:20140216T090000')),

    # today until tomorrow
    ('22:00  1:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20140216T220000',
         'DTEND;TZID=Europe/Berlin:20140217T010000')),

    # other timezone
    ('22:00 1:00 Europe/London Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/London:20140216T220000',
         'DTEND;TZID=Europe/London:20140217T010000')),

    ('15.06. Äwesöme Event',
     _create_vevent('DTSTART;VALUE=DATE:20140615',
                    'DTEND;VALUE=DATE:20140616')),
)


@freeze_time('2014-02-16 12:00:00')
def test__construct_event_format_de_complexer():
    for data_list, vevent in test_set_format_de_complexer:
        event = _construct_event(data_list.split(), locale=LOCALE_BERLIN)
        assert _replace_uid(event).to_ical() == vevent


test_set_leap_year = _create_testcases(
    ('29.02. Äwesöme Event',
     _create_vevent(
         'DTSTART;VALUE=DATE:20160229',
         'DTEND;VALUE=DATE:20160301',
         'DTSTAMP:20160101T202122Z')),
)


def test_leap_year():
    for data_list, vevent in test_set_leap_year:
        with freeze_time('1999-1-1'):
            with pytest.raises(DateTimeParseError):
                event = _construct_event(data_list.split(), locale=LOCALE_BERLIN)
        with freeze_time('2016-1-1 20:21:22'):
            event = _construct_event(data_list.split(), locale=LOCALE_BERLIN)
            assert _replace_uid(event).to_ical() == vevent


test_set_description = _create_testcases(
    # now events where the start date has to be inferred, too
    # today
    ('8:00 Äwesöme Event :: this is going to be awesome',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20140216T080000',
         'DTEND;TZID=Europe/Berlin:20140216T090000',
         'DESCRIPTION:this is going to be awesome')),

    # today until tomorrow
    ('22:00  1:00 Äwesöme Event :: Will be even better',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20140216T220000',
         'DTEND;TZID=Europe/Berlin:20140217T010000',
         'DESCRIPTION:Will be even better')),

    ('15.06. Äwesöme Event :: and again',
     _create_vevent('DTSTART;VALUE=DATE:20140615',
                    'DTEND;VALUE=DATE:20140616',
                    'DESCRIPTION:and again')),
)


def test_description():
    for data_list, vevent in test_set_description:
        with freeze_time('2014-02-16 12:00:00'):
            event = _construct_event(data_list.split(), locale=LOCALE_BERLIN)
            assert _replace_uid(event).to_ical() == vevent


test_set_repeat_floating = _create_testcases(
    # now events where the start date has to be inferred, too
    # today
    ('8:00 Äwesöme Event',
     _create_vevent(
         'DTSTART:20140216T080000',
         'DTEND:20140216T090000',
         'DESCRIPTION:please describe the event',
         'RRULE:FREQ=DAILY;UNTIL=20150604T000000')))


def test_repeat_floating():
    for data_list, vevent in test_set_repeat_floating:
        with freeze_time('2014-02-16 12:00:00'):
            event = _construct_event(data_list.split(),
                                     description='please describe the event',
                                     repeat='daily',
                                     until='04.06.2015',
                                     locale=LOCALE_FLOATING)
            assert normalize_component(_replace_uid(event).to_ical()) == \
                normalize_component(vevent)


test_set_repeat_localized = _create_testcases(
    # now events where the start date has to be inferred, too
    # today
    ('8:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20140216T080000',
         'DTEND;TZID=Europe/Berlin:20140216T090000',
         'DESCRIPTION:please describe the event',
         'RRULE:FREQ=DAILY;UNTIL=20150604T230000Z')))


def test_repeat_localized():
    for data_list, vevent in test_set_repeat_localized:
        with freeze_time('2014-02-16 12:00:00'):
            event = _construct_event(data_list.split(),
                                     description='please describe the event',
                                     repeat='daily',
                                     until='05.06.2015',
                                     locale=LOCALE_BERLIN)
            assert normalize_component(_replace_uid(event).to_ical()) == \
                normalize_component(vevent)


test_set_alarm = _create_testcases(
    ('8:00 Äwesöme Event',
     ['BEGIN:VEVENT',
      'SUMMARY:Äwesöme Event',
      'DTSTART;TZID=Europe/Berlin:20140216T080000',
      'DTEND;TZID=Europe/Berlin:20140216T090000',
      'DTSTAMP:20140216T120000Z',
      'UID:E41JRQX2DB4P1AQZI86BAT7NHPBHPRIIHQKA',
      'DESCRIPTION:please describe the event',
      'BEGIN:VALARM',
      'ACTION:DISPLAY',
      'DESCRIPTION:please describe the event',
      'TRIGGER:-PT23M',
      'END:VALARM',
      'END:VEVENT']))


@freeze_time('2014-02-16 12:00:00')
def test_alarm():
    for data_list, vevent in test_set_alarm:
        event = _construct_event(data_list.split(),
                                 description='please describe the event',
                                 alarm='23m',
                                 locale=LOCALE_BERLIN)
        assert _replace_uid(event).to_ical() == vevent


test_set_description_and_location_and_categories = _create_testcases(
    # now events where the start date has to be inferred, too
    # today
    ('8:00 Äwesöme Event',
     _create_vevent(
         'DTSTART;TZID=Europe/Berlin:20140216T080000',
         'DTEND;TZID=Europe/Berlin:20140216T090000',
         'CATEGORIES:boring meeting',
         'DESCRIPTION:please describe the event',
         'LOCATION:in the office')))


@freeze_time('2014-02-16 12:00:00')
def test_description_and_location_and_categories():
    for data_list, vevent in test_set_description_and_location_and_categories:
        event = _construct_event(data_list.split(),
                                 description='please describe the event',
                                 location='in the office',
                                 categories=['boring meeting'],
                                 locale=LOCALE_BERLIN)
        assert _replace_uid(event).to_ical() == vevent



from __future__ import annotations

import io

import click


class CanvasTranslator:
    """Translates a canvas object into a printable string."""

    def __init__(self, canvas, palette: dict[str, str] | None = None) -> None:
        """currently only support foreground colors, so palette is
        a dictionary of attributes and foreground colors"""
        self._canvas = canvas
        self._palette : dict[str, tuple[bool, str]]= {}
        if palette:
            for key, color in palette.items():
                self.add_color(key, color)

    def add_color(self, key: str, color: str) -> None:
        if color.startswith('#'):  # RGB colour
            r = color[1:3]
            g = color[3:5]
            b = color[5:8]
            rgb = int(r, 16), int(g, 16), int(b, 16)
            value = True, '\33[38;2;{!s};{!s};{!s}m'.format(*rgb)
        else:
            color = color.split(' ')[-1]
            if color == 'gray':
                color = 'white'  # click will insist on US-english
            value = False, color

        self._palette[key] = value  # (is_ansi, color)

    def transform(self) -> str:
        self.output = io.StringIO()
        for row in self._canvas.content():
            # self.spaces = 0
            for col in row[:-1]:
                self._process_char(*col)
            # the last column has all the trailing whitespace, which deforms
            # everything if the terminal is resized:
            col = row[-1]
            self._process_char(col[0], col[1], col[2].rstrip())

            self.output.write('\n')

        return self.output.getvalue()

    def _process_char(self, fmt, _, b):
        text = b.decode()
        if not fmt:
            self.output.write(text)
        else:
            fmt = self._palette[fmt]
            if fmt[0]:
                self.output.write(f'{fmt[1]}{click.style(text)}')
            else:
                self.output.write(click.style(text, fg=fmt[1]))



import datetime as dt
from textwrap import dedent

import pytest
from freezegun import freeze_time

from khal import exceptions
from khal.controllers import import_ics, khal_list, start_end_from_daterange
from khal.khalendar.vdir import Item

from . import utils
from .utils import _get_text

today = dt.date.today()
yesterday = today - dt.timedelta(days=1)
tomorrow = today + dt.timedelta(days=1)

event_allday_template = """BEGIN:VEVENT
SEQUENCE:0
UID:uid3@host1.com
DTSTART;VALUE=DATE:{}
DTEND;VALUE=DATE:{}
SUMMARY:a meeting
DESCRIPTION:short description
LOCATION:LDB Lobby
END:VEVENT"""

event_today = event_allday_template.format(today.strftime('%Y%m%d'),
                                           tomorrow.strftime('%Y%m%d'))
item_today = Item(event_today)

event_format = '{calendar-color}{start-end-time-style:16} {title}'
event_format += '{repeat-symbol}{description-separator}{description}{calendar-color}'

conf = {'locale': utils.LOCALE_BERLIN,
        'default': {'timedelta': dt.timedelta(days=2), 'show_all_days': False}
        }


class TestGetAgenda:
    def test_new_event(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        event = coll.create_event_from_ics(event_today, utils.cal1)
        coll.insert(event)
        assert ['                 a meeting :: short description\x1b[0m'] == \
            khal_list(coll, [], conf, agenda_format=event_format, day_format="")

    def test_new_event_day_format(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        event = coll.create_event_from_ics(event_today, utils.cal1)
        coll.insert(event)
        assert ['Today\x1b[0m',
                '                 a meeting :: short description\x1b[0m'] == \
            khal_list(coll, [], conf, agenda_format=event_format, day_format="{name}")

    def test_agenda_default_day_format(self, coll_vdirs):
        with freeze_time('2016-04-10 12:33'):
            today = dt.date.today()
            event_today = event_allday_template.format(
                today.strftime('%Y%m%d'), tomorrow.strftime('%Y%m%d'))
            coll, vdirs = coll_vdirs
            event = coll.create_event_from_ics(event_today, utils.cal1)
            coll.insert(event)
            out = khal_list(
                coll, conf=conf, agenda_format=event_format, datepoint=[])
            assert [
                '\x1b[1m10.04.2016 12:33\x1b[0m\x1b[0m',
                '↦                a meeting :: short description\x1b[0m'] == out

    def test_agenda_fail(self, coll_vdirs):
        with freeze_time('2016-04-10 12:33'):
            coll, vdirs = coll_vdirs
            with pytest.raises(exceptions.FatalError):
                khal_list(coll, conf=conf, agenda_format=event_format, datepoint=['xyz'])
            with pytest.raises(exceptions.FatalError):
                khal_list(coll, conf=conf, agenda_format=event_format, datepoint=['today'])

    def test_empty_recurrence(self, coll_vdirs):
        coll, vidrs = coll_vdirs
        coll.insert(coll.create_event_from_ics(dedent(
            'BEGIN:VEVENT\r\n'
            'UID:no_recurrences\r\n'
            'SUMMARY:No recurrences\r\n'
            'RRULE:FREQ=DAILY;COUNT=2;INTERVAL=1\r\n'
            'EXDATE:20110908T130000\r\n'
            'EXDATE:20110909T130000\r\n'
            'DTSTART:20110908T130000\r\n'
            'DTEND:20110908T170000\r\n'
            'END:VEVENT\r\n'
        ), utils.cal1))
        assert '\n'.join(khal_list(coll, [], conf,
                         agenda_format=event_format, day_format="{name}")).lower() == ''


class TestImport:
    def test_import(self, coll_vdirs):
        coll, vdirs = coll_vdirs
        view = {'event_format': '{title}'}
        conf = {'locale': utils.LOCALE_BERLIN, 'view': view}
        import_ics(coll, conf, _get_text('event_rrule_recuid'), batch=True)
        start_date = utils.BERLIN.localize(dt.datetime(2014, 4, 30))
        end_date = utils.BERLIN.localize(dt.datetime(2014, 9, 26))
        events = list(coll.get_localized(start_date, end_date))
        assert len(events) == 6
        events = sorted(events)
        assert events[1].start_local == utils.BERLIN.localize(dt.datetime(2014, 7, 7, 9, 0))
        assert utils.BERLIN.localize(dt.datetime(2014, 7, 14, 7, 0)) in \
            [ev.start for ev in events]

        import_ics(coll, conf, _get_text('event_rrule_recuid_update'), batch=True)
        events = list(coll.get_localized(start_date, end_date))
        for ev in events:
            print(ev.start)
            assert ev.calendar == 'foobar'
        assert len(events) == 5
        assert utils.BERLIN.localize(dt.datetime(2014, 7, 14, 7, 0)) not in \
            [ev.start_local for ev in events]

    def test_mix_datetime_types(self, coll_vdirs):
        """
        Test importing events with mixed tz-aware and tz-naive datetimes.
        """
        coll, vdirs = coll_vdirs
        view = {'event_format': '{title}'}
        import_ics(
            coll,
            {'locale': utils.LOCALE_BERLIN, 'view': view},
            _get_text('event_dt_mixed_awareness'),
            batch=True
        )
        start_date = utils.BERLIN.localize(dt.datetime(2015, 5, 29))
        end_date = utils.BERLIN.localize(dt.datetime(2015, 6, 3))
        events = list(coll.get_localized(start_date, end_date))
        assert len(events) == 2
        events = sorted(events)
        assert events[0].start_local == \
            utils.BERLIN.localize(dt.datetime(2015, 5, 30, 12, 0))
        assert events[0].end_local == \
            utils.BERLIN.localize(dt.datetime(2015, 5, 30, 16, 0))
        assert events[1].start_local == \
            utils.BERLIN.localize(dt.datetime(2015, 6, 2, 12, 0))
        assert events[1].end_local == \
            utils.BERLIN.localize(dt.datetime(2015, 6, 2, 16, 0))


def test_start_end():
    with freeze_time('2016-04-10'):
        start = dt.datetime(2016, 4, 10, 0, 0)
        end = dt.datetime(2016, 4, 11, 0, 0)
        assert (start, end) == start_end_from_daterange(('today',), locale=utils.LOCALE_BERLIN)


def test_start_end_default_delta():
    with freeze_time('2016-04-10'):
        start = dt.datetime(2016, 4, 10, 0, 0)
        end = dt.datetime(2016, 4, 11, 0, 0)
        assert (start, end) == start_end_from_daterange(('today',), utils.LOCALE_BERLIN)


def test_start_end_delta():
    with freeze_time('2016-04-10'):
        start = dt.datetime(2016, 4, 10, 0, 0)
        end = dt.datetime(2016, 4, 12, 0, 0)
        assert (start, end) == start_end_from_daterange(('today', '2d'), utils.LOCALE_BERLIN)


def test_start_end_empty():
    with freeze_time('2016-04-10'):
        start = dt.datetime(2016, 4, 10, 0, 0)
        end = dt.datetime(2016, 4, 11, 0, 0)
        assert (start, end) == start_end_from_daterange([], utils.LOCALE_BERLIN)


def test_start_end_empty_default():
    with freeze_time('2016-04-10'):
        start = dt.datetime(2016, 4, 10, 0, 0)
        end = dt.datetime(2016, 4, 13, 0, 0)
        assert (start, end) == start_end_from_daterange(
            [], utils.LOCALE_BERLIN,
            default_timedelta_date=dt.timedelta(days=3),
            default_timedelta_datetime=dt.timedelta(hours=1),
        )



import datetime as dt

from freezegun import freeze_time

from khal.ui.calendarwidget import CalendarWidget

on_press: dict = {}

keybindings = {
    'today': ['T'],
    'left': ['left', 'h', 'backspace'],
    'up': ['up', 'k'],
    'right': ['right', 'l', ' '],
    'down': ['down', 'j'],
}


def test_initial_focus_today():
    today = dt.date.today()
    frame = CalendarWidget(on_date_change=lambda _: None,
                           keybindings=keybindings,
                           on_press=on_press,
                           weeknumbers='right')
    assert frame.focus_date == today


def test_set_focus_date():
    today = dt.date.today()
    for diff in range(-10, 10, 1):
        frame = CalendarWidget(on_date_change=lambda _: None,
                               keybindings=keybindings,
                               on_press=on_press,
                               weeknumbers='right')
        day = today + dt.timedelta(days=diff)
        frame.set_focus_date(day)
        assert frame.focus_date == day


def test_set_focus_date_weekstart_6():

    with freeze_time('2016-04-10'):
        today = dt.date.today()
        for diff in range(-21, 21, 1):
            frame = CalendarWidget(on_date_change=lambda _: None,
                                   keybindings=keybindings,
                                   on_press=on_press,
                                   firstweekday=6,
                                   weeknumbers='right')
            day = today + dt.timedelta(days=diff)
            frame.set_focus_date(day)
            assert frame.focus_date == day

    with freeze_time('2016-04-23'):
        today = dt.date.today()
        for diff in range(10):
            frame = CalendarWidget(on_date_change=lambda _: None,
                                   keybindings=keybindings,
                                   on_press=on_press,
                                   firstweekday=6,
                                   weeknumbers='right')
            day = today + dt.timedelta(days=diff)
            frame.set_focus_date(day)
            assert frame.focus_date == day


def test_set_focus_far_future():
    future_date = dt.date.today() + dt.timedelta(days=1000)
    frame = CalendarWidget(on_date_change=lambda _: None,
                           keybindings=keybindings,
                           on_press=on_press,
                           weeknumbers='right')
    frame.set_focus_date(future_date)
    assert frame.focus_date == future_date

def test_set_focus_far_past():
    future_date = dt.date.today() - dt.timedelta(days=1000)
    frame = CalendarWidget(on_date_change=lambda _: None,
                           keybindings=keybindings,
                           on_press=on_press,
                           weeknumbers='right')
    frame.set_focus_date(future_date)
    assert frame.focus_date == future_date



import datetime as dt
import os.path

import pytest
from tzlocal import get_localzone as _get_localzone

from khal.settings import get_config
from khal.settings.exceptions import CannotParseConfigFileError, InvalidSettingsError
from khal.settings.utils import (
    config_checks,
    get_all_vdirs,
    get_color_from_vdir,
    get_unique_name,
    is_color,
)

try:
    # Available from configobj 5.1.0
    from configobj.validate import VdtValueError
except ModuleNotFoundError:
    from validate import VdtValueError

from .utils import LOCALE_BERLIN

PATH = __file__.rsplit('/', 1)[0] + '/configs/'


def get_localzone():
    # this reproduces the code in settings.util for the time being
    import pytz
    return pytz.timezone(str(_get_localzone()))


class TestSettings:
    def test_simple_config(self):
        config = get_config(
            PATH + 'simple.conf',
            _get_color_from_vdir=lambda x: None,
            _get_vdir_type=lambda x: 'calendar',
        )
        comp_config = {
            'calendars': {
                'home': {
                    'path': os.path.expanduser('~/.calendars/home/'), 'readonly': False,
                    'color': None, 'priority': 10, 'type': 'calendar', 'addresses': [''],
                },
                'work': {
                    'path': os.path.expanduser('~/.calendars/work/'), 'readonly': False,
                    'color': None, 'priority': 10, 'type': 'calendar', 'addresses': [''],
                },
            },
            'sqlite': {'path': os.path.expanduser('~/.local/share/khal/khal.db')},
            'locale': LOCALE_BERLIN,
            'default': {
                'default_calendar': None,
                'print_new': 'False',
                'highlight_event_days': False,
                'timedelta': dt.timedelta(days=2),
                'default_event_duration': dt.timedelta(hours=1),
                'default_dayevent_duration': dt.timedelta(days=1),
                'default_event_alarm': dt.timedelta(0),
                'default_dayevent_alarm': dt.timedelta(0),
                'show_all_days': False,
                'enable_mouse': True,
            }
        }
        for key in comp_config:
            assert config[key] == comp_config[key]

    def test_nocalendars(self):
        with pytest.raises(InvalidSettingsError):
            get_config(PATH + 'nocalendars.conf')

    def test_one_level_calendar(self):
        with pytest.raises(InvalidSettingsError):
            get_config(PATH + 'one_level_calendars.conf')

    def test_small(self):
        config = get_config(
            PATH + 'small.conf',
            _get_color_from_vdir=lambda x: None,
            _get_vdir_type=lambda x: 'calendar',
        )
        comp_config = {
            'calendars': {
                'home': {'path': os.path.expanduser('~/.calendars/home/'),
                         'color': 'dark green', 'readonly': False, 'priority': 20,
                         'type': 'calendar', 'addresses': ['']},
                'work': {'path': os.path.expanduser('~/.calendars/work/'),
                         'readonly': True, 'color': None, 'priority': 10,
                         'type': 'calendar', 'addresses': ['user@example.com']}},
            'sqlite': {'path': os.path.expanduser('~/.local/share/khal/khal.db')},
            'locale': {
                'local_timezone': get_localzone(),
                'default_timezone': get_localzone(),
                'timeformat': '%X',
                'dateformat': '%x',
                'longdateformat': '%x',
                'datetimeformat': '%c',
                'longdatetimeformat': '%c',
                'firstweekday': 0,
                'unicode_symbols': True,
                'weeknumbers': False,
            },
            'default': {
                'default_calendar': None,
                'print_new': 'False',
                'highlight_event_days': False,
                'timedelta': dt.timedelta(days=2),
                'default_event_duration': dt.timedelta(hours=1),
                'default_dayevent_duration': dt.timedelta(days=1),
                'show_all_days': False,
                'enable_mouse': True,
                'default_event_alarm': dt.timedelta(0),
                'default_dayevent_alarm': dt.timedelta(0),
            }
        }
        for key in comp_config:
            assert config[key] == comp_config[key]

    def test_old_config(self, tmpdir):
        old_config = """
[Calendar home]
path: ~/.khal/calendars/home/
color: dark blue
[sqlite]
path: ~/.khal/khal.db
[locale]
timeformat: %H:%M
dateformat: %d.%m.
longdateformat: %d.%m.%Y
[default]
"""
        conf_path = str(tmpdir.join('old.conf'))
        with open(conf_path, 'w+') as conf:
            conf.write(old_config)
        with pytest.raises(CannotParseConfigFileError):
            get_config(conf_path)

    def test_extra_sections(self, tmpdir):
        config = """
[calendars]
[[home]]
path = ~/.khal/calendars/home/
color = dark blue
unknown = 42
[unknownsection]
foo = bar
"""
        conf_path = str(tmpdir.join('old.conf'))
        with open(conf_path, 'w+') as conf:
            conf.write(config)
        get_config(conf_path)
        # FIXME test for log entries

    def test_default_calendar_readonly(self, tmpdir):
        config = """
[calendars]
[[home]]
path = ~/.khal/calendars/home/
color = dark blue
readonly = True
[default]
default_calendar = home
"""
        conf_path = str(tmpdir.join('old.conf'))
        with open(conf_path, 'w+') as conf:
            conf.write(config)
        with pytest.raises(InvalidSettingsError):
            config_checks(get_config(conf_path))


def test_broken_color(metavdirs):
    path = metavdirs
    newvdir = path + '/cal5/'
    os.makedirs(newvdir)
    with open(newvdir + 'color', 'w') as metafile:
        metafile.write('xxx')
    assert get_color_from_vdir(newvdir) is None


def test_discover(metavdirs):
    test_vdirs = {
        '/cal1/public', '/cal1/private', '/cal2/public',
        '/cal3/home', '/cal3/public', '/cal3/work',
        '/cal4/cfgcolor', '/cal4/dircolor', '/cal4/cfgcolor_again',
        '/cal4/cfgcolor_once_more',
        '/singlecollection',
    }
    path = metavdirs
    assert test_vdirs == {vdir[len(path):] for vdir in get_all_vdirs(path + '/**/*/')}
    assert test_vdirs == {vdir[len(path):] for vdir in get_all_vdirs(path + '/**/')}
    assert test_vdirs == {vdir[len(path):] for vdir in get_all_vdirs(path + '/**/*')}


def test_get_unique_name(metavdirs):
    path = metavdirs
    vdirs = list(get_all_vdirs(path + '/**/'))
    names = []
    for vdir in sorted(vdirs):
        names.append(get_unique_name(vdir, names))
    assert sorted(names) == sorted([
        'my private calendar', 'my calendar', 'public', 'home', 'public1',
        'work', 'cfgcolor', 'cfgcolor_again', 'cfgcolor_once_more', 'dircolor',
        'singlecollection',
    ])


def test_config_checks(metavdirs):
    path = metavdirs
    config = {
        'calendars': {
            'default': {'path': path + '/cal[1-3]/*', 'type': 'discover'},
            'calendars_color': {'path': path + '/cal4/*', 'type': 'discover', 'color': 'dark blue'},
        },
        'sqlite': {'path': '/tmp'},
        'locale': {'default_timezone': 'Europe/Berlin', 'local_timezone': 'Europe/Berlin'},
        'default': {'default_calendar': None},
    }
    config_checks(config)
    # cut off the part of the path that changes on each run
    for cal in config['calendars']:
        config['calendars'][cal]['path'] = config['calendars'][cal]['path'][len(metavdirs):]

    test_config = {
        'calendars': {
            'home': {
                'color': None,
                'path': '/cal3/home',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'my calendar': {
                'color': 'dark blue',
                'path': '/cal1/public',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'my private calendar': {
                'color': '#FF00FF',
                'path': '/cal1/private',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'public1': {
                'color': None,
                'path': '/cal3/public',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'public': {
                'color': None,
                'path': '/cal2/public',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'work': {
                'color': None,
                'path': '/cal3/work',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'cfgcolor': {
                'color': 'dark blue',
                'path': '/cal4/cfgcolor',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'dircolor': {
                'color': 'dark blue',
                'path': '/cal4/dircolor',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'cfgcolor_again': {
                'color': 'dark blue',
                'path': '/cal4/cfgcolor_again',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },
            'cfgcolor_once_more': {
                'color': 'dark blue',
                'path': '/cal4/cfgcolor_once_more',
                'readonly': False,
                'type': 'calendar',
                'priority': 10,
            },

        },
        'default': {'default_calendar': None},
        'locale': {'default_timezone': 'Europe/Berlin', 'local_timezone': 'Europe/Berlin'},
        'sqlite': {'path': '/tmp'},
    }

    assert config['calendars'] == test_config['calendars']
    assert config == test_config


def test_is_color():
    assert is_color('dark blue') == 'dark blue'
    assert is_color('#123456') == '#123456'
    assert is_color('123') == '123'
    with pytest.raises(VdtValueError):
        assert is_color('red') == 'red'



import datetime as dt
import locale
import platform
import unicodedata

import pytest

from khal.calendar_display import (
    get_calendar_color,
    get_color_list,
    getweeknumber,
    str_week,
    vertical_month,
)

today = dt.date.today()
yesterday = today - dt.timedelta(days=1)
tomorrow = today + dt.timedelta(days=1)


def test_getweeknumber():
    assert getweeknumber(dt.date(2011, 12, 12)) == 50
    assert getweeknumber(dt.date(2011, 12, 31)) == 52
    assert getweeknumber(dt.date(2012, 1, 1)) == 52
    assert getweeknumber(dt.date(2012, 1, 2)) == 1


def test_str_week():
    aday = dt.date(2012, 6, 1)
    bday = dt.date(2012, 6, 8)
    week = [dt.date(2012, 6, 6),
            dt.date(2012, 6, 7),
            dt.date(2012, 6, 8),
            dt.date(2012, 6, 9),
            dt.date(2012, 6, 10),
            dt.date(2012, 6, 11),
            dt.date(2012, 6, 12),
            dt.date(2012, 6, 13)]
    assert str_week(week, aday) == ' 6  7  8  9 10 11 12 13 '
    assert str_week(week, bday) == ' 6  7 \x1b[7m 8\x1b[0m  9 10 11 12 13 '


class testCollection:
    def __init__(self) -> None:
        self._calendars : dict[str, dict]= {}

    def addCalendar(self, name: str , color: str, priority: int) -> None:
        self._calendars[name] = {'color': color, 'priority': priority}


def test_get_calendar_color():

    exampleCollection = testCollection()
    exampleCollection.addCalendar('testCalendar1', 'dark red', 20)
    exampleCollection.addCalendar('testCalendar2', 'light green', 10)
    exampleCollection.addCalendar('testCalendar3', '', 10)

    assert get_calendar_color('testCalendar1', 'light blue', exampleCollection) == 'dark red'
    assert get_calendar_color('testCalendar2', 'light blue', exampleCollection) == 'light green'

    # test default color
    assert get_calendar_color('testCalendar3', 'light blue', exampleCollection) == 'light blue'


def test_get_color_list():

    exampleCalendarList = ['testCalendar1', 'testCalendar2']

    # test different priorities
    exampleCollection1 = testCollection()
    exampleCollection1.addCalendar('testCalendar1', 'dark red', 20)
    exampleCollection1.addCalendar('testCalendar2', 'light green', 10)

    testList1 = get_color_list(exampleCalendarList, 'light_blue', exampleCollection1)
    assert 'dark red' in testList1
    assert len(testList1) == 1

    # test same priorities
    exampleCollection2 = testCollection()
    exampleCollection2.addCalendar('testCalendar1', 'dark red', 20)
    exampleCollection2.addCalendar('testCalendar2', 'light green', 20)

    testList2 = get_color_list(exampleCalendarList, 'light_blue', exampleCollection2)
    assert 'dark red' in testList2
    assert 'light green' in testList2
    assert len(testList2) == 2

    # test duplicated colors
    exampleCollection3 = testCollection()
    exampleCollection3.addCalendar('testCalendar1', 'dark red', 20)
    exampleCollection3.addCalendar('testCalendar2', 'dark red', 20)

    testList3 = get_color_list(exampleCalendarList, 'light_blue', exampleCollection3)
    assert len(testList3) == 1

    # test indexing operator (required by str_highlight_day())
    exampleCollection4 = testCollection()
    exampleCollection4.addCalendar('testCalendar1', 'dark red', 20)
    exampleCollection4.addCalendar('testCalendar2', 'dark red', 20)

    testList3 = get_color_list(exampleCalendarList, 'light_blue', exampleCollection4)
    assert testList3[0] == 'dark red'


example1 = [
    '\x1b[1m    Mo Tu We Th Fr Sa Su \x1b[0m',
    '\x1b[1mDec \x1b[0m28 29 30  1  2  3  4 ',
    '     5  6  7  8  9 10 11 ',
    '    \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '    19 20 21 22 23 24 25 ',
    '\x1b[1mJan \x1b[0m26 27 28 29 30 31  1 ',
    '     2  3  4  5  6  7  8 ',
    '     9 10 11 12 13 14 15 ',
    '    16 17 18 19 20 21 22 ',
    '    23 24 25 26 27 28 29 ',
    '\x1b[1mFeb \x1b[0m30 31  1  2  3  4  5 ',
    '     6  7  8  9 10 11 12 ',
    '    13 14 15 16 17 18 19 ',
    '    20 21 22 23 24 25 26 ',
    '\x1b[1mMar \x1b[0m27 28 29  1  2  3  4 ']

example2 = [
    '\x1b[1m    Mo Tu We Th Fr Sa Su \x1b[0m',
    '    28 29 30  1  2  3  4 ',
    '\x1b[1mDec \x1b[0m 5  6  7  8  9 10 11 ',
    '    \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '    19 20 21 22 23 24 25 ',
    '    26 27 28 29 30 31  1 ',
    '\x1b[1mJan \x1b[0m 2  3  4  5  6  7  8 ',
    '     9 10 11 12 13 14 15 ',
    '    16 17 18 19 20 21 22 ',
    '    23 24 25 26 27 28 29 ',
    '    30 31  1  2  3  4  5 ',
    '\x1b[1mFeb \x1b[0m 6  7  8  9 10 11 12 ',
    '    13 14 15 16 17 18 19 ',
    '    20 21 22 23 24 25 26 ',
    '    27 28 29  1  2  3  4 ']


example_weno = [
    '\x1b[1m    Mo Tu We Th Fr Sa Su   \x1b[0m',
    '\x1b[1mDec \x1b[0m28 29 30  1  2  3  4 \x1b[1m48\x1b[0m',
    '     5  6  7  8  9 10 11 \x1b[1m49\x1b[0m',
    '    \x1b[7m12\x1b[0m 13 14 15 16 17 18 \x1b[1m50\x1b[0m',
    '    19 20 21 22 23 24 25 \x1b[1m51\x1b[0m',
    '\x1b[1mJan \x1b[0m26 27 28 29 30 31  1 \x1b[1m52\x1b[0m',
    '     2  3  4  5  6  7  8 \x1b[1m 1\x1b[0m',
    '     9 10 11 12 13 14 15 \x1b[1m 2\x1b[0m',
    '    16 17 18 19 20 21 22 \x1b[1m 3\x1b[0m',
    '    23 24 25 26 27 28 29 \x1b[1m 4\x1b[0m',
    '\x1b[1mFeb \x1b[0m30 31  1  2  3  4  5 \x1b[1m 5\x1b[0m',
    '     6  7  8  9 10 11 12 \x1b[1m 6\x1b[0m',
    '    13 14 15 16 17 18 19 \x1b[1m 7\x1b[0m',
    '    20 21 22 23 24 25 26 \x1b[1m 8\x1b[0m',
    '\x1b[1mMar \x1b[0m27 28 29  1  2  3  4 \x1b[1m 9\x1b[0m']

example_we_start_su = [
    '\x1b[1m    Su Mo Tu We Th Fr Sa \x1b[0m',
    '\x1b[1mDec \x1b[0m27 28 29 30  1  2  3 ',
    '     4  5  6  7  8  9 10 ',
    '    11 \x1b[7m12\x1b[0m 13 14 15 16 17 ',
    '    18 19 20 21 22 23 24 ',
    '    25 26 27 28 29 30 31 ',
    '\x1b[1mJan \x1b[0m 1  2  3  4  5  6  7 ',
    '     8  9 10 11 12 13 14 ',
    '    15 16 17 18 19 20 21 ',
    '    22 23 24 25 26 27 28 ',
    '\x1b[1mFeb \x1b[0m29 30 31  1  2  3  4 ',
    '     5  6  7  8  9 10 11 ',
    '    12 13 14 15 16 17 18 ',
    '    19 20 21 22 23 24 25 ',
    '\x1b[1mMar \x1b[0m26 27 28 29  1  2  3 ']

example_cz = [
    '\x1b[1m    Po \xdat St \u010ct P\xe1 So Ne \x1b[0m',
    '\x1b[1mpro \x1b[0m28 29 30  1  2  3  4 ',
    '     5  6  7  8  9 10 11 ',
    '    \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '    19 20 21 22 23 24 25 ',
    '\x1b[1mled \x1b[0m26 27 28 29 30 31  1 ',
    '     2  3  4  5  6  7  8 ',
    '     9 10 11 12 13 14 15 ',
    '    16 17 18 19 20 21 22 ',
    '    23 24 25 26 27 28 29 ',
    '\x1b[1m\xfano \x1b[0m30 31  1  2  3  4  5 ',
    '     6  7  8  9 10 11 12 ',
    '    13 14 15 16 17 18 19 ',
    '    20 21 22 23 24 25 26 ',
    '\x1b[1mb\u0159e \x1b[0m27 28 29  1  2  3  4 ']

example_gr = [
    '\x1b[1m     δε τρ τε πε πα σα κυ \x1b[0m',
    '\x1b[1mδεκ  \x1b[0m28 29 30  1  2  3  4 ',
    '      5  6  7  8  9 10 11 ',
    '     \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '     19 20 21 22 23 24 25 ',
    '\x1b[1mιαν  \x1b[0m26 27 28 29 30 31  1 ',
    '      2  3  4  5  6  7  8 ',
    '      9 10 11 12 13 14 15 ',
    '     16 17 18 19 20 21 22 ',
    '     23 24 25 26 27 28 29 ',
    '\x1b[1mφεβ  \x1b[0m30 31  1  2  3  4  5 ',
    '      6  7  8  9 10 11 12 ',
    '     13 14 15 16 17 18 19 ',
    '     20 21 22 23 24 25 26 ',
    '\x1b[1mμαρ  \x1b[0m27 28 29  1  2  3  4 ']

example_gr_darwin = [
    '\x1b[1m    δε τρ τε πε πα σα κυ \x1b[0m',
    '\x1b[1mδεκ \x1b[0m28 29 30  1  2  3  4 ',
    '     5  6  7  8  9 10 11 ',
    '    \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '    19 20 21 22 23 24 25 ',
    '\x1b[1mιαν \x1b[0m26 27 28 29 30 31  1 ',
    '     2  3  4  5  6  7  8 ',
    '     9 10 11 12 13 14 15 ',
    '    16 17 18 19 20 21 22 ',
    '    23 24 25 26 27 28 29 ',
    '\x1b[1mφεβ \x1b[0m30 31  1  2  3  4  5 ',
    '     6  7  8  9 10 11 12 ',
    '    13 14 15 16 17 18 19 ',
    '    20 21 22 23 24 25 26 ',
    '\x1b[1mμαρ \x1b[0m27 28 29  1  2  3  4 ']

example_de = [
    '\x1b[1m    Mo Di Mi Do Fr Sa So \x1b[0m',
    '\x1b[1mDez \x1b[0m28 29 30  1  2  3  4 ',
    '     5  6  7  8  9 10 11 ',
    '    \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '    19 20 21 22 23 24 25 ',
    '\x1b[1mJan \x1b[0m26 27 28 29 30 31  1 ',
    '     2  3  4  5  6  7  8 ',
    '     9 10 11 12 13 14 15 ',
    '    16 17 18 19 20 21 22 ',
    '    23 24 25 26 27 28 29 ',
    '\x1b[1mFeb \x1b[0m30 31  1  2  3  4  5 ',
    '     6  7  8  9 10 11 12 ',
    '    13 14 15 16 17 18 19 ',
    '    20 21 22 23 24 25 26 ',
    '\x1b[1mMär \x1b[0m27 28 29  1  2  3  4 ']

example_de_freebsd = [
    '\x1b[1m     Mo Di Mi Do Fr Sa So \x1b[0m',
    '\x1b[1mDez. \x1b[0m28 29 30  1  2  3  4 ',
    '      5  6  7  8  9 10 11 ',
    '     \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '     19 20 21 22 23 24 25 ',
    '\x1b[1mJan. \x1b[0m26 27 28 29 30 31  1 ',
    '      2  3  4  5  6  7  8 ',
    '      9 10 11 12 13 14 15 ',
    '     16 17 18 19 20 21 22 ',
    '     23 24 25 26 27 28 29 ',
    '\x1b[1mFeb. \x1b[0m30 31  1  2  3  4  5 ',
    '      6  7  8  9 10 11 12 ',
    '     13 14 15 16 17 18 19 ',
    '     20 21 22 23 24 25 26 ',
    '\x1b[1mMärz \x1b[0m27 28 29  1  2  3  4 ']

example_de_netbsd = [
    '\x1b[1m     Mo Di Mi Do Fr Sa So \x1b[0m',
    '\x1b[1mDez. \x1b[0m28 29 30  1  2  3  4 ',
    '      5  6  7  8  9 10 11 ',
    '     \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '     19 20 21 22 23 24 25 ',
    '\x1b[1mJan. \x1b[0m26 27 28 29 30 31  1 ',
    '      2  3  4  5  6  7  8 ',
    '      9 10 11 12 13 14 15 ',
    '     16 17 18 19 20 21 22 ',
    '     23 24 25 26 27 28 29 ',
    '\x1b[1mFeb. \x1b[0m30 31  1  2  3  4  5 ',
    '      6  7  8  9 10 11 12 ',
    '     13 14 15 16 17 18 19 ',
    '     20 21 22 23 24 25 26 ',
    '\x1b[1mM\xe4r. \x1b[0m27 28 29  1  2  3  4 ']

example_fr = [
    '\x1b[1m      lu ma me je ve sa di \x1b[0m',
    '\x1b[1mdéc.  \x1b[0m28 29 30  1  2  3  4 ',
    '       5  6  7  8  9 10 11 ',
    '      \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '      19 20 21 22 23 24 25 ',
    '\x1b[1mjanv. \x1b[0m26 27 28 29 30 31  1 ',
    '       2  3  4  5  6  7  8 ',
    '       9 10 11 12 13 14 15 ',
    '      16 17 18 19 20 21 22 ',
    '      23 24 25 26 27 28 29 ',
    '\x1b[1mfévr. \x1b[0m30 31  1  2  3  4  5 ',
    '       6  7  8  9 10 11 12 ',
    '      13 14 15 16 17 18 19 ',
    '      20 21 22 23 24 25 26 ',
    '\x1b[1mmars  \x1b[0m27 28 29  1  2  3  4 ']

example_fr_darwin = [
    '\x1b[1m    Lu Ma Me Je Ve Sa Di \x1b[0m',
    '\x1b[1mdéc \x1b[0m28 29 30  1  2  3  4 ',
    '     5  6  7  8  9 10 11 ',
    '    \x1b[7m12\x1b[0m 13 14 15 16 17 18 ',
    '    19 20 21 22 23 24 25 ',
    '\x1b[1mjan \x1b[0m26 27 28 29 30 31  1 ',
    '     2  3  4  5  6  7  8 ',
    '     9 10 11 12 13 14 15 ',
    '    16 17 18 19 20 21 22 ',
    '    23 24 25 26 27 28 29 ',
    '\x1b[1mfév \x1b[0m30 31  1  2  3  4  5 ',
    '     6  7  8  9 10 11 12 ',
    '    13 14 15 16 17 18 19 ',
    '    20 21 22 23 24 25 26 ',
    '\x1b[1mmar \x1b[0m27 28 29  1  2  3  4 ']


def test_vertical_month():
    try:
        locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')
        vert_str = vertical_month(month=12, year=2011,
                                  today=dt.date(2011, 12, 12))
        assert vert_str == example1

        vert_str = vertical_month(month=12, year=2011,
                                  today=dt.date(2011, 12, 12),
                                  monthdisplay='firstfullweek')
        assert vert_str == example2

        weno_str = vertical_month(month=12, year=2011,
                                  today=dt.date(2011, 12, 12),
                                  weeknumber='right')
        assert weno_str == example_weno

        we_start_su_str = vertical_month(
            month=12, year=2011,
            today=dt.date(2011, 12, 12),
            firstweekday=6)
        assert we_start_su_str == example_we_start_su
    except locale.Error as error:
        if str(error) == 'unsupported locale setting':
            pytest.xfail(
                'To get this test to run, you need to add `en_US.utf-8` to '
                'your locales. On Debian GNU/Linux 8 you do this by '
                'uncommenting `de_DE.utf-8 in /etc/locale.gen and then run '
                '`locale-gen` (as root).'
            )
    finally:
        locale.setlocale(locale.LC_ALL, 'C')


def test_vertical_month_unicode():
    try:
        locale.setlocale(locale.LC_ALL, 'de_DE.UTF-8')
        vert_str = vertical_month(month=12, year=2011,
                                  today=dt.date(2011, 12, 12))
        # de_DE locale on at least Net and FreeBSD is different from the one
        # commonly used on linux systems
        if platform.system() == 'FreeBSD':
            assert vert_str == example_de_freebsd
        elif platform.system() == 'NetBSD':
            assert vert_str == example_de_netbsd
        else:
            assert vert_str == example_de

        '\n'.join(vert_str)  # issue 142
    except locale.Error as error:
        if str(error) == 'unsupported locale setting':
            pytest.xfail(
                'To get this test to run, you need to add `de_DE.utf-8` to '
                'your locales. On Debian GNU/Linux 8 you do this by '
                'uncommenting `de_DE.utf-8 in /etc/locale.gen and then run '
                '`locale-gen` (as root).'
            )
        else:
            raise
    finally:
        locale.setlocale(locale.LC_ALL, 'C')


def test_vertical_month_unicode_weekdeays():
    try:
        locale.setlocale(locale.LC_ALL, 'cs_CZ.UTF-8')
        vert_str = vertical_month(month=12, year=2011,
                                  today=dt.date(2011, 12, 12))
        assert [line.lower() for line in vert_str] == [line.lower() for line in example_cz]
        '\n'.join(vert_str)  # issue 142/293
    except locale.Error as error:
        if str(error) == 'unsupported locale setting':
            pytest.xfail(
                'To get this test to run, you need to add `cs_CZ.UTF-8` to '
                'your locales. On Debian GNU/Linux 8 you do this by '
                'uncommenting `cs_CZ.UTF-8` in /etc/locale.gen and then run '
                '`locale-gen` (as root).'
            )
        else:
            raise
    finally:
        locale.setlocale(locale.LC_ALL, 'C')


def strip_accents(string):
    """remove accents from unicode characters"""
    return ''.join(c for c in unicodedata.normalize('NFD', string)
                   if unicodedata.category(c) != 'Mn')


def test_vertical_month_unicode_weekdeays_gr():
    try:
        locale.setlocale(locale.LC_ALL, 'el_GR.UTF-8')
        vert_str = vertical_month(month=12, year=2011,
                                  today=dt.date(2011, 12, 12))
        # on some OSes, Greek locale's abbreviated day of the week and
        # month names have accents, on some they haven't
        if platform.system() == 'Darwin':
            assert strip_accents('\n'.join([line.lower() for line in vert_str])) == \
                '\n'.join(example_gr_darwin)
        else:
            assert strip_accents('\n'.join([line.lower() for line in vert_str])) == \
                '\n'.join(example_gr)
        '\n'.join(vert_str)  # issue 142/293
    except locale.Error as error:
        if str(error) == 'unsupported locale setting':
            pytest.xfail(
                'To get this test to run, you need to add `el_GR.UTF-8` to '
                'your locales. On Debian GNU/Linux 8 you do this by '
                'uncommenting `el_GR.UTF-8` in /etc/locale.gen and then run '
                '`locale-gen` (as root).'
            )
        else:
            raise
    finally:
        locale.setlocale(locale.LC_ALL, 'C')


def test_vertical_month_abbr_fr():
    # see issue #653
    try:
        locale.setlocale(locale.LC_ALL, 'fr_FR.UTF-8')
        vert_str = vertical_month(month=12, year=2011,
                                  today=dt.date(2011, 12, 12))
        if platform.system() == 'Darwin':
            assert '\n'.join(vert_str) == '\n'.join(example_fr_darwin)
        else:
            assert '\n'.join(vert_str) == '\n'.join(example_fr)
    except locale.Error as error:
        if str(error) == 'unsupported locale setting':
            pytest.xfail(
                'To get this test to run, you need to add `fr_FR.UTF-8` to '
                'your locales. On Debian GNU/Linux 8 you do this by '
                'uncommenting `fr_FR.UTF-8` in /etc/locale.gen and then run '
                '`locale-gen` (as root).'
            )
        else:
            raise
    finally:
        locale.setlocale(locale.LC_ALL, 'C')
